/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import math.VecMath;

import common.IConverter;

public final class Circle<T extends Number> implements IShape<T> {

	private final VecMath<T> mt;
	private T[] ctr;
	private T r;
	private T inv_r;
	private T sq_r;
	private T inv_2rrr;
	private T[] tmp;
	public Circle(T[] center,T r,VecMath<T> math) {
		this.mt = math;
		this.ctr = center;
		setRadius(r);
		this.tmp = math.util.array(center.length);
	}
	private Circle(T[] center,T r,T invr,T inv2rrr,T sqr,VecMath<T> math) {
		this.mt = math;
		this.ctr = center;
		this.r = r;
		this.inv_r = invr;
		this.inv_2rrr = inv2rrr;
		this.sq_r = sqr;
		this.tmp = math.util.array(center.length);
	}
	@Override
	public T[] getCenter() {
		return ctr;
	}
	@Override
	public synchronized boolean reflect(Ray<T> r1,Ray<T> r2) {
		// Float test to decide if high precision calculation is necessary
		float tmp0 = r1.getStart()[0].floatValue()-ctr[0].floatValue();
		float tmp1 = r1.getStart()[1].floatValue()-ctr[1].floatValue();
		float d = r1.getDir()[1].floatValue()*tmp0-r1.getDir()[0].floatValue()*tmp1;
		if(Math.abs(d)<r.floatValue()+0.0001f) {
			// High precision calculation
			mt.diff(r1.getStart(),ctr,tmp);
			T proj = mt.dot(r1.getDir(),tmp);
			if(mt.cmp(proj,mt.ZERO)>0)
				return false;
			mt.scaleAdd(tmp,mt.neg(proj),r1.getDir(),tmp);
			T tmp2 = mt.normSquared(tmp);
			if(mt.cmp(tmp2,sq_r)<0) {
				int p = mt.util.minPrecision(r1.getStart());
				mt.scaleAdd(tmp,mt.neg(mt.sqrt(mt.sub(sq_r,tmp2),p)),r1.getDir(),tmp);
				mt.add(ctr,tmp,r2.getStart());
				T dn = mt.sub(mt.normSquared(tmp),sq_r);
				if(mt.util.getScale(dn)+2<mt.util.getScale(sq_r))
					mt.scale(mt.sub(inv_r,mt.mul(inv_2rrr,dn)),tmp);
				else
					mt.normalize(tmp);
				mt.mirror(r1.getDir(),tmp,r2.getDir());
				return true;
			}
		}
		r2.copy(r1);
		return false;
	}
	private T sq(T t) {
		return mt.mul(t,t);
	}
	public T getRadius() {
		return r;
	}
	public void setRadius(T radius) {
		this.r = radius;
		this.inv_r = mt.inv(radius);
		this.sq_r = sq(radius);
		this.inv_2rrr = mt.inv(mt.mul(mt.mul(sq_r,r),mt.TWO));
	}
	@Override
	public void setCenter(T[] c) {
		this.ctr = c;
	}
	@Override
	public Circle<T> copy() {
		return new Circle<T>(ctr.clone(),r,inv_r,inv_2rrr,sq_r,mt);
	}
	@Override
	public void rotate(T[] p,T[] dir,T phi) {
		mt.rotate2D(ctr,p,phi);
	}
	@Override
	public void rotate(T[] p,T[] dir,T sinphi,T cosphi) {
		mt.rotate2D(ctr,p,sinphi,cosphi);
	}
	@Override
	public void scale(T s) {
		setRadius(mt.mul(r,s));
	}
	@Override
	public <S extends Number>Circle<S> convert(IConverter<T,S> cv,VecMath<S> smath) {
		return new Circle<S>(cv.convert(ctr),cv.convert(r),smath);
	}
	@Override
	public T getSize() {
		return r;
	}
	@Override
	public void setSize(T r) {
		setRadius(r);
	}
	@Override
	public synchronized boolean contains(T[] p) {
		mt.diff(ctr,p,tmp);
		return mt.cmp(mt.normSquared(tmp),sq_r)<=0;
	}
}
