/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import math.Util;
import math.VecMath;

import common.IConverter;

public class Lattice<T extends Number> extends BoxAdapter<T> {

	// Temporary variables that are expensive to create
	private final LinkedList<IBox<T>> tempCells;
	// Grid points per direction
	private short points;
	// Gap between two grid points
	private T gap;
	// Gap^-1 for optimization
	private T invGap;
	// Cell centers
	private T[][] localCellCenters;
	// Cell corners
	private T[][] localCellCorners;
	// 2 or 3 dimensional
	protected final byte dim;
	// Contains all math operations
	protected final VecMath<T> math;
	// Shortcut for math.util
	protected final Util<T> util;
	public Lattice(T[] p,short sidePoints,T gap,VecMath<T> math) {
		this.math = math;
		this.util = math.util;
		T sideLength = math.mul(gap,sidePoints-1);
		this.box = new Box2D<T>(p,sideLength,math);
		this.points = sidePoints;
		this.gap = gap;
		this.invGap = math.inv(gap);
		this.tempCells = new LinkedList<IBox<T>>();
		this.dim = (byte)p.length;
		init();
	}
	protected Lattice(IBox<T> box,short sidePoints,T gap,VecMath<T> math) {
		this.math = math;
		this.util = math.util;
		this.box = box;
		this.points = sidePoints;
		this.gap = gap;
		this.invGap = math.inv(gap);
		this.tempCells = new LinkedList<IBox<T>>();
		this.dim = (byte)box.getCorner().length;
		init();
	}
	private void init() {
		localCellCenters = util.array((short)Math.round(Math.pow(points-1,dim)),dim);
		localCellCorners = util.array((short)Math.round(Math.pow(points-1,dim)),dim);
		IndexIterator idxIter = new IndexIterator();
		T[] diagHalf = math.getDiagHalf(dim).clone();
		math.scale(gap,diagHalf);
		while(idxIter.hasNext()) {
			short[] idxs = idxIter.next();
			short idx = getTableIndex(idxs);
			localCellCorners[idx] = util.array(dim);
			math.linBaseComb(idxs,math.getBase(dim),localCellCorners[idx]);
			math.scale(gap,localCellCorners[idx]);
			localCellCenters[idx] = util.array(dim);
			math.add(localCellCorners[idx],diagHalf,localCellCenters[idx]);
		}
	}
	private short getTableIndex(short[] idxs) {
		short idx = 0;
		for(int i = 0;i<idxs.length;i++)
			idx += idxs[i]*Math.round(Math.pow(points-1,i));
		return idx;
	}
	private void getIndex(T[] localPoint,short[] idxVec) {
		// clip intersection point p to lattice
		for(int i = 0;i<idxVec.length;i++) {
			idxVec[i] = math.mul(localPoint[i],invGap).shortValue();
			idxVec[i] = idxVec[i]<0?0:(short)Math.min(idxVec[i],points-2);
		}
	}
	public void intersect(Ray<T> ray,List<T[]> pts) {
		LocalCellIterator iter = new LocalCellIterator();
		iter.init(ray,null);
		while(iter.hasNext()) {
			T[] tmp = util.array(dim);
			local2WorldPoint(iter.next(),tmp);
			pts.add(tmp);
		}
	}
	public boolean getCellCenter(short[] idxs,T[] cellCtr) {
		if(checkRange(idxs)) {
			short[] dIdxs = new short[idxs.length];
			for(byte i = 0;i<dIdxs.length;i++)
				dIdxs[i] = 1;
			math.linBaseComb(dIdxs,getBaseVectors(),cellCtr);
			math.scale(math.ONE_OVER_TWO,cellCtr);
			math.linBaseCombAdd(idxs,getBaseVectors(),cellCtr);
			math.scaleAdd(getCorner(),gap,cellCtr,cellCtr);
			return true;
		}
		return false;
	}
	protected void getCellCenterLocal(short[] idxs,T[] cellCtr) {
		util.copy(localCellCenters[getTableIndex(idxs)],cellCtr);
	}
	public void getCellCenter(T[] worldPoint,T[] cellCtr) {
		short[] idxs = new short[dim];
		world2LocalPoint(worldPoint,cellCtr);
		getIndex(cellCtr,idxs);
		getCellCenter(idxs,cellCtr);
	}
	protected void getCellCenterLocal(T[] localPoint,T[] cellCtr) {
		short[] idxs = new short[dim];
		getIndex(localPoint,idxs);
		getCellCenterLocal(idxs,cellCtr);
	}
	public short getPoints() {
		return points;
	}
	public void setPoints(short pts) {
		this.points = pts;
		this.tempCells.clear();
		init();
	}
	public T getGap() {
		return gap;
	}
	public void setGap(T gp) {
		this.gap = gp;
		this.invGap = math.inv(gap);
		this.tempCells.clear();
		init();
	}
	public void setBound(IBox<T> box) {
		this.box = box;
		this.tempCells.clear();
	}
	private boolean checkRange(short[] idxs) {
		for(int i = 0;i<idxs.length;i++)
			if(0>idxs[i]||points-1<=idxs[i])
				return false;
		return true;
	}
	final boolean calcStartCellIndex(Ray<T> localRay,short[] idxs,T[] tmp) {
		boolean ret = true;
		if(containsLocal(localRay.getStart()))
			getIndex(localRay.getStart(),idxs);
		else {
			ret = localIntersect(localRay,tmp);
			if(ret)
				getIndex(tmp,idxs);
		}
		return ret;
	}
	final boolean calcNextCellIndex(Ray<T> locRay,short[] idxs,T[] tmp) {
		short[] diagIdxs = idxs.clone();
		IBox<T> tmpCell = pollTmpCell();
		for(int steps = 0;steps<dim-1;steps++) {
			for(byte j = 0;j<dim;j++) {
				short dIdx = (short)(math.cmp(locRay.getDir()[j],math.ZERO)<0?-1:1);
				if(steps==1)
					dIdx = (short)(-dIdx);
				short[] tmpIdxs = idxs.clone();
				tmpIdxs[j] += dIdx;
				if(checkRange(tmpIdxs)) {
					getCellCenterLocal(tmpIdxs,tmp);
					math.diff(tmp,locRay.getStart(),tmp);
					T t = math.dot(tmp,locRay.getDir());
					math.scaleAdd(locRay.getStart(),t,locRay.getDir(),tmp);
					tmpCell.setCorner(localCellCorners[getTableIndex(tmpIdxs)]);
					math.diff(tmp,tmpCell.getCorner(),tmp);
					if(tmpCell.containsLocal(tmp)) {
						System.arraycopy(tmpIdxs,0,idxs,0,idxs.length);
						pushTmpCell(tmpCell);
						return true;
					}
				}
				if(steps==0)
					diagIdxs[j] += dIdx;
			}
			System.arraycopy(diagIdxs,0,idxs,0,idxs.length);
		}
		pushTmpCell(tmpCell);
		return checkRange(idxs);
	}
	final boolean calcNextCellIndexDouble(Ray<T> locRay,short[] idxs,double[] tmp) {
		short[] diagIdxs = idxs.clone();
		IBox<T> tmpCell = pollTmpCell();
		for(int steps = 0;steps<dim-1;steps++) {
			for(byte j = 0;j<dim;j++) {
				short dIdx = (short)(math.cmp(locRay.getDir()[j],math.ZERO)<0?-1:1);
				if(steps==1)
					dIdx = (short)(-dIdx);
				short[] tmpIdxs = idxs.clone();
				tmpIdxs[j] += dIdx;
				if(checkRange(tmpIdxs)) {
					T[] cellCtr = localCellCenters[getTableIndex(tmpIdxs)];
					double[] start = math.util.cast2doubleArray(locRay.getStart());
					double[] dir = math.util.cast2doubleArray(locRay.getDir());
					for(int i = 0;i<dim;i++)
						tmp[i] = cellCtr[i].doubleValue()-start[i];
					double t = 0;
					for(int i = 0;i<dim;i++)
						t += tmp[i]*dir[i];
					for(int i = 0;i<dim;i++)
						tmp[i] = start[i]+t*dir[i];
					tmpCell.setCorner(localCellCorners[getTableIndex(tmpIdxs)]);
					for(int i = 0;i<dim;i++)
						tmp[i] = tmp[i]-tmpCell.getCorner()[i].doubleValue();
					if(tmpCell.containsLocal(math.util.cast(tmp))) {
						System.arraycopy(tmpIdxs,0,idxs,0,idxs.length);
						pushTmpCell(tmpCell);
						return true;
					}
				}
				if(steps==0)
					diagIdxs[j] += dIdx;
			}
			System.arraycopy(diagIdxs,0,idxs,0,idxs.length);
		}
		pushTmpCell(tmpCell);
		return checkRange(idxs);
	}
	private IBox<T> pollTmpCell() {
		IBox<T> tmpCell;
		if(tempCells.size()>0)
			tmpCell = tempCells.pollFirst();
		else {
			tmpCell = box.copy();
			tmpCell.setLength(gap);
		}
		return tmpCell;
	}
	private void pushTmpCell(IBox<T> b) {
		tempCells.add(b);
	}
	@Override
	public void rotate(T[] p,T[] ax,T phi) {
		super.rotate(p,ax,phi);
		tempCells.clear();
	}
	@Override
	public void rotate(T[] p,T[] ax,T sinphi,T cosphi) {
		super.rotate(p,ax,sinphi,cosphi);
		tempCells.clear();
	}
	@Override
	public <S extends Number>Lattice<S> convert(IConverter<T,S> cv,VecMath<S> smath) {
		IBox<S> sbox = super.convert(cv,smath);
		return new Lattice<S>(sbox,points,cv.convert(gap),smath);
	}
	protected class IndexIterator implements Iterator<short[]> {

		private short[] idxs;
		private short pts;
		public IndexIterator() {
			this.idxs = new short[dim];
			this.idxs[0] = -1;
			this.pts = points;
		}
		@Override
		public boolean hasNext() {
			for(int i = 0;i<idxs.length;i++)
				if(idxs[i]<pts-2)
					return true;
			return false;
		}
		@Override
		public short[] next() {
			if(hasNext()) {
				for(int i = 0;i<idxs.length;i++)
					if(idxs[i]==pts-2)
						idxs[i] = 0;
					else {
						++idxs[i];
						break;
					}
				return idxs;
			}
			return null;
		}
		@Override
		public void remove() {}
	}
	protected final class LocalCellIterator implements Iterator<T[]> {

		private final short[] idxs;
		private final T[] tmp1;
		private final double[] tmp2;
		private short[] startIdx;
		private Ray<T> localRay;
		private Ray<T> lowPrecLocalRay;
		private boolean empty = false;
		private boolean next = false;
		private int marginScale = Integer.MIN_VALUE;
		private boolean useLowPrecRay = false;
		public LocalCellIterator() {
			this.tmp1 = util.array(dim);
			this.tmp2 = new double[dim];
			this.idxs = new short[dim];
		}
		public boolean hasNext() {
			if(empty)
				return false;
			if(!next) {
				if(-marginScale>10)
					next = calcNextCellIndex(getCellFinderRay(),idxs,tmp1);
				else
					next = calcNextCellIndexDouble(getCellFinderRay(),idxs,tmp2);
				empty = !next;
			}
			return !empty;
		}
		public boolean init(Ray<T> ray,int mScale,short[] startCellIdx) {
			this.marginScale = mScale;
			return init(ray,startCellIdx);
		}
		public boolean init(Ray<T> ray,short[] startCellIdx) {
			if(localRay==null) {
				T[] start = util.array(dim);
				T[] dir = util.array(dim);
				localRay = new Ray<T>(start,dir,util);
			}
			world2LocalPoint(ray.getStart(),localRay.getStart());
			world2LocalVec(ray.getDir(),localRay.getDir());
			convert2LowPrecRay();
			if(!(empty = startCellIdx==null))
				System.arraycopy(startCellIdx,0,idxs,0,idxs.length);
			else
				empty = !calcStartCellIndex(getCellFinderRay(),idxs,tmp1);
			if(!empty) {
				next = true;
				startIdx = idxs.clone();
			}
			return !empty;
		}
		public short[] getStartCellIndex() {
			return startIdx;
		}
		private Ray<T> getCellFinderRay() {
			return useLowPrecRay?lowPrecLocalRay:localRay;
		}
		private void convert2LowPrecRay() {
			if(marginScale>=Util.MIN_INT+10) {
				int prec = -marginScale+10;
				useLowPrecRay = prec<util.maxPrecision(localRay.getDir());
				if(useLowPrecRay) {
					if(lowPrecLocalRay==null) {
						T[] start = util.array(dim);
						T[] dir = util.array(dim);
						lowPrecLocalRay = new Ray<T>(start,dir,util);
					}
					T[] start = lowPrecLocalRay.getStart();
					start[0] = util.copy(localRay.getStart()[0],prec);
					start[1] = util.copy(localRay.getStart()[1],prec);
					T[] dir = lowPrecLocalRay.getDir();
					dir[0] = util.copy(localRay.getDir()[0],prec);
					dir[1] = util.copy(localRay.getDir()[1],prec);
				}
			}
		}
		public T[] next() {
			T[] pt = util.array(dim);
			boolean res = next(pt);
			if(!res)
				pt = null;
			return pt;
		}
		public void remove() {}
		public void continueWithRay(Ray<T> locRay) {
			this.localRay = locRay;
			convert2LowPrecRay();
		}
		public Ray<T> getLocalRay() {
			return localRay;
		}
		public void setMarginScale(short masc) {
			this.marginScale = masc;
			convert2LowPrecRay();
		}
		public boolean next(T[] pt) {
			if(!empty)
				if(next||hasNext()) {
					next = false;
					getCellCenterLocal(idxs,pt);
					return true;
				}
			return false;
		}
		/**
		 * The next call to next() gives the same value as the previously
		 * called next().
		 */
		public void replay() {
			next = true;
		}
	}
}
