/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package lattice;
import math.Util;
import math.VecMath;

import common.IConverter;
import common.Log;

public final class Ray<T extends Number> {

	private T[] start;
	private T[] dir;
	private final Util<T> util;
	private Object attr;
	public Ray(T[] start,T[] dir,Util<T> util) {
		this.start = start;
		this.dir = dir;
		this.util = util;
		checkDirNorm();
	}
	public Ray(Ray<T> r,int p,Util<T> util) {
		this.start = r.getStart().clone();
		this.dir = r.getDir().clone();
		this.util = util;
		if(p!=-1)
			setPrecision(p);
		checkDirNorm();
	}
	public Ray(Util<T> util) {
		this.util = util;
	}
	public T[] getDir() {
		return dir;
	}
	public T[] getStart() {
		return start;
	}
	public void setStart(T[] start) {
		this.start = start;
	}
	public void copy(Ray<T> r) {
		if(r.start!=null)
			this.start = r.start.clone();
		else
			this.start = null;
		if(r.dir!=null)
			this.dir = r.dir.clone();
		else
			this.dir = null;
	}
	public void setPrecision(int p) {
		for(int i = 0;i<start.length;i++) {
			if(start[i]!=null)
				start[i] = util.copy(start[i],p);
			if(dir[i]!=null)
				dir[i] = util.copy(dir[i],p);
		}
	}
	public void getPoint(T param,T[] retVal,VecMath<T> math) {
		math.scaleAdd(start,param,dir,retVal);
	}
	private void checkDirNorm() {
		float dirNormSq = 0;
		for(int i = 0;i<dir.length;i++) {
			if(dir[i]==null)
				return;
			dirNormSq += dir[i].floatValue()*dir[i].floatValue();
		}
		if(Math.abs(dirNormSq-1)>0.00001f)
			Log.getIstc().logln("ERROR: Dir vector of ray not normalized");
	}
	public String toString() {
		String s = "Ray:\n";
		s += "start 0:"+start[0]+"\n";
		s += "start 1:"+start[1]+"\n";
		s += "dir 0:"+dir[0]+"\n";
		s += "dir 1:"+dir[1]+"\n";
		return s;
	}
	public int getScale() {
		int es0 = util.getScale(start[0]);
		int es1 = util.getScale(start[1]);
		int es2 = util.getScale(dir[0]);
		int es3 = util.getScale(dir[1]);
		return Math.max(Math.max(es0,es1),Math.max(es2,es3));
	}
	/**
	 * Returns the number of digits behind the comma.
	 */
	public int getAccuracy() {
		int es0 = util.getErrorScale(start[0]);
		int es1 = util.getErrorScale(start[1]);
		int es2 = util.getErrorScale(dir[0]);
		int es3 = util.getErrorScale(dir[1]);
		return -Math.max(Math.max(es0,es1),Math.max(es2,es3))-1;
	}
	public void incPrecision(int p) {
		for(int i = 0;i<start.length;i++) {
			if(start[i]!=null) {
				int p0 = util.precision(start[i]);
				if(p0<Integer.MAX_VALUE-p)
					start[i] = util.copy(start[i],p0+p);
			}
			if(dir[i]!=null) {
				int p0 = util.precision(dir[i]);
				if(p0<Integer.MAX_VALUE-p)
					dir[i] = util.copy(dir[i],p0+p);
			}
		}
		// Log.getIstc().logln("Inc ray ac: " + ac0 + "->" + getAccuracy());
	}
	public <S extends Number>Ray<S> convert(IConverter<T,S> c,Util<S> ut) {
		return new Ray<S>(c.convert(start),c.convert(dir),ut);
	}
	public void setAttribute(Object o){ this.attr = o; }
	public Object getAttribute() { return attr; }
}