/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package math;
import java.util.HashMap;

import common.Globals;
import common.Log;

public abstract class VecMath<T extends Number> implements IBasicOperations<T> {

	public final T ZERO;
	public final T ONE;
	public final T MINUS_ONE;
	public final T TWO;
	public final T THREE;
	public final T FOUR;
	public final T FIVE;
	public final T TEN;
	public final T HUNDRED;
	public final T THOUSAND;
	public final T MINUS_TEN;
	public final T ONE_OVER_TWO;
	public final T ONE_OVER_TEN;
	public final T ONE_OVER_HUNDRED;
	public final T ONE_OVER_THOUSAND;
	public final T[] BASE3D_X;
	public final T[] BASE3D_Y;
	public final T[] BASE3D_Z;
	public final T[][] BASE3D;
	public final T[] BASE2D_X;
	public final T[] BASE2D_Y;
	public final T[][] BASE2D;
	public final T[] ZERO2D;
	public final T[] DIAG_HALF2D;
	public final T[] DIAG_HALF3D;
	public final T[] ZERO3D;
	// Upper precision bound for all calculations
	public int PRECISION;
	public T EPSILON;
	public T EPSILON10;
	public T PI;
	public T PI_TIMES_TWO;
	public T PI_OVER_TWO;
	public T PI_OVER_FOUR;
	public T PI_OVER_EIGHT;
	public T E;
	public T[] sinTab;
	public final Util<T> util;
	private int tablePrec;
	private boolean tables;
	private boolean log;
	public VecMath(Util<T> util,int precision,boolean tables,boolean lg) {
		this.PRECISION = precision;
		this.util = util;
		this.tablePrec = -1;
		this.tables = tables;
		this.log = lg;
		ZERO = util.cast(0);
		ONE = util.cast(1);
		MINUS_ONE = util.cast(-1);
		TWO = util.cast(2);
		THREE = util.cast(3);
		FOUR = util.cast(4);
		FIVE = util.cast(5);
		TEN = util.cast(10);
		MINUS_TEN = util.cast(-10);
		HUNDRED = util.cast(100);
		THOUSAND = util.cast(1000);
		ONE_OVER_TWO = div(ONE,TWO);
		ONE_OVER_TEN = div(ONE,TEN);
		ONE_OVER_HUNDRED = div(ONE,HUNDRED);
		ONE_OVER_THOUSAND = div(ONE,THOUSAND);
		BASE3D_X = util.array(3);
		BASE3D_X[0] = ONE;
		BASE3D_X[1] = ZERO;
		BASE3D_X[2] = ZERO;
		BASE3D_Y = util.array(3);
		BASE3D_Y[0] = ZERO;
		BASE3D_Y[1] = ONE;
		BASE3D_Y[2] = ZERO;
		BASE3D_Z = util.array(3);
		BASE3D_Z[0] = ZERO;
		BASE3D_Z[1] = ZERO;
		BASE3D_Z[2] = ONE;
		BASE3D = util.array(3,3);
		BASE3D[0] = BASE3D_X;
		BASE3D[1] = BASE3D_Y;
		BASE3D[2] = BASE3D_Z;
		BASE2D_X = util.array(2);
		BASE2D_X[0] = ONE;
		BASE2D_X[1] = ZERO;
		BASE2D_Y = util.array(2);
		BASE2D_Y[0] = ZERO;
		BASE2D_Y[1] = ONE;
		BASE2D = util.array(2,2);
		BASE2D[0] = BASE2D_X;
		BASE2D[1] = BASE2D_Y;
		ZERO2D = util.array(2);
		ZERO2D[0] = ZERO;
		ZERO2D[1] = ZERO;
		DIAG_HALF2D = util.array(2);
		DIAG_HALF2D[0] = ONE_OVER_TWO;
		DIAG_HALF2D[1] = ONE_OVER_TWO;
		DIAG_HALF3D = util.array(3);
		DIAG_HALF3D[0] = ONE_OVER_TWO;
		DIAG_HALF3D[1] = ONE_OVER_TWO;
		DIAG_HALF3D[2] = ONE_OVER_TWO;
		ZERO3D = util.array(3);
		ZERO3D[0] = ZERO;
		ZERO3D[1] = ZERO;
		ZERO3D[2] = ZERO;
		calcPrecisionDependentValues();
	}
	public void setPrecision(int precision) {
		PRECISION = precision;
		calcPrecisionDependentValues();
	}
	public void enableTables(){
		tables = true;
		updateTables();
	}
	private void calcPrecisionDependentValues() {
		long t = 0;
		if(log){
			Log.getIstc().logln("Calculating pi & e in precision "+PRECISION+"...");
			t = System.currentTimeMillis();
		}
		PI = util.getPi(PRECISION);
		E = util.getE(PRECISION);
		if(log){
			t -= System.currentTimeMillis();
			Log.getIstc().logln("... pi & e finished ("+(-t)+" ms)");
		}
		PI_TIMES_TWO = mul(PI,TWO);
		PI_OVER_TWO = div(PI,TWO);
		PI_OVER_FOUR = div(PI,FOUR);
		PI_OVER_EIGHT = div(PI_OVER_FOUR,TWO);
		EPSILON = pow(ONE_OVER_TEN,PRECISION);
		EPSILON10 = pow(ONE_OVER_TEN,PRECISION-10);
		updateTables();
	}
	private void updateTables(){
		if(tables&&(tablePrec<PRECISION)){
			Log.getIstc().logln("Calculating sine tables in precision "+PRECISION+"...");
			// the discrSinCos(...) method assumes for the array size l: l%2 = 0
			long t = System.currentTimeMillis();
			tablePrec = PRECISION;
			sinTab = util.array(1000);
			calcSinTab(sinTab);
			t -= System.currentTimeMillis();
			Log.getIstc().logln("... sine table finished ("+(-t)+" ms)");
		}
	}
	private void calcSinTab(T[] sin_tab) {
		short digits = (short)Math.log10(sinTab.length);
		T[][] tmpSin = util.array(digits,10);
		T[][] tmpCos = util.array(digits,10);
		tmpSin[digits-1][1] = sin(div(PI,THOUSAND));
		for(int k = tmpSin.length-1;k>=0;k--) {
			tmpSin[k][0] = ZERO;
			tmpCos[k][0] = ONE;
			for(int i = 2;i<tmpSin[k].length;i++) {
				tmpCos[k][i-1] = sqrt(sub(ONE,mul(tmpSin[k][i-1],tmpSin[k][i-1])));
				if(k==0&&i>5)
					tmpCos[k][i-1] = neg(tmpCos[k][i-1]);
				tmpSin[k][i] = add(mul(tmpSin[k][i-1],tmpCos[k][1]),
				        mul(tmpCos[k][i-1],tmpSin[k][1]));
			}
			tmpCos[k][9] = sqrt(sub(ONE,mul(tmpSin[k][9],tmpSin[k][9])));
			if(k>0)
				tmpSin[k-1][1] = add(mul(tmpSin[k][9],tmpCos[k][1]),mul(tmpCos[k][9],tmpSin[k][1]));
		}
		for(int i = 0;i<10;i++)
			sin_tab[i] = tmpSin[digits-1][i];
		HashMap<Integer,T> cosHash = new HashMap<Integer,T>();
		for(int i = 10;i<sin_tab.length;i++) {
			String iStr = i+"";
			int num1 = Integer.parseInt(iStr.substring(0,1));
			int num2 = Integer.parseInt(iStr.substring(1));
			sin_tab[i] = tmpSin[digits-iStr.length()][num1];
			int fullNum = (int)Math.round(num1*Math.pow(10,iStr.length()-1));
			T cos1 = cosHash.get(fullNum);
			if(cos1==null) {
				cos1 = sqrt(sub(ONE,mul(sin_tab[i],sin_tab[i])));
				if(i>sin_tab.length/2)
					cos1 = neg(cos1);
				cosHash.put(fullNum,cos1);
			}
			T cos2 = cosHash.get(num2);
			if(cos2==null) {
				cos2 = sqrt(sub(ONE,mul(sin_tab[num2],sin_tab[num2])));
				cosHash.put(num2,cos2);
			}
			sin_tab[i] = add(mul(sin_tab[i],cos2),mul(cos1,sin_tab[num2]));
		}
	}
	/**
	 * Returns the discrete angle for the given pi factor
	 */
	public T discrAngle(float pifactor) {
		boolean neg = false;
		if(pifactor<0){ 
			pifactor = -pifactor;
			neg = true;
		}
		int idx = Math.min((int)(pifactor*sinTab.length+0.5f),sinTab.length-1);
		T t1 = util.cast(idx);
		T t2 = util.cast(sinTab.length);
		T retVal = mul(PI,div(t1,t2));
		if(neg)
			retVal = neg(retVal);
		return retVal;
	}
	/**
	 * retrieves precise sin and cos values for discrete angles 
	 * phi in a fast way.
	 * phi = pi*((int)(pifactor*sinTab.length))/sinTab.length 
	 * 
	 * abs(pifactor) < 1
	 */
	public T[] discrSinCos(float pifactor) {
		T[] ret = util.array(2);
		boolean neg = false;
		if(pifactor<0) {
			neg = true;
			pifactor = -pifactor;
		}
		if(pifactor<1) {
			int idx = Math.min((int)(pifactor*sinTab.length+0.5f),sinTab.length-1);
			ret[0] = sinTab[idx];
			if(neg)
				ret[0] = neg(ret[0]);
			if(idx>=sinTab.length/2) {
				neg = true;
				idx -= sinTab.length/2;
			} else {
				neg = false;
				idx = sinTab.length/2-idx;
			}
			ret[1] = sinTab[idx];
			if(neg)
				ret[1] = neg(ret[1]);
		} else
			Log.getIstc().logln("Error: Sine table not defined for abs(phi) >= pi");
		return ret;
	}
	public T[] sincos(T phi) {
		T[] ret = null;
		boolean neg = false;
		if(cmp(phi,ZERO)<0) {
			neg = true;
			phi = neg(phi);
		}
		if(cmp(phi,PI)>0) {
			T k = util.cast(div(sub(phi,PI),PI_TIMES_TWO).intValue());
			phi = sub(phi,mul(add(k,ONE),PI_TIMES_TWO));
		}
		if(cmp(phi,ZERO)<0) {
			neg = !neg;
			phi = neg(phi);
		}
		ret = util.array(2);
		ret[0] = sin(phi);
		ret[1] = sqrt(sub(ONE,mul(ret[0],ret[0])));
		if(cmp(phi,PI_OVER_TWO)>0)
			ret[1] = neg(ret[1]);
		if(neg)
			ret[0] = neg(ret[0]);
		return ret;
	}
	public T[] getBase(byte idx,byte dim) {
		return dim==3?BASE3D[idx]:BASE2D[idx];
	}
	public T[][] getBase(byte dim) {
		return dim==3?BASE3D:BASE2D;
	}
	public T[] getDiagHalf(byte dim) {
		return dim==3?DIAG_HALF3D:DIAG_HALF2D;
	}
	public void swap(T[] vec,byte idx1,byte idx2) {
		T tmp = vec[idx1];
		vec[idx1] = vec[idx2];
		vec[idx2] = tmp;
	}
	public T norm(T[] vec) {
		return sqrt(normSquared(vec));
	}
	public T dot(T[] vec1,T[] vec2) {
		T retVal = ZERO;
		for(int i = 0;i<vec1.length;i++)
			retVal = add(retVal,mul(vec1[i],vec2[i]));
		return retVal;
	}
	public void normalize(T[] vec) {
		T norm = norm(vec);
		for(int i = 0;i<vec.length;i++)
			vec[i] = div(vec[i],norm);
	}
	public void diff(T[] vec1,T[] vec2,T[] vec3) {
		for(int i = 0;i<vec1.length;i++)
			vec3[i] = sub(vec1[i],vec2[i]);
	}
	public void scaleAdd(T[] v1,T x,T[] v2,T[] v3) {
		for(int i = 0;i<v1.length;i++)
			v3[i] = add(v1[i],mul(x,v2[i]));
	}
	public T normSquared(T[] vec) {
		T retVal = ZERO;
		for(int i = 0;i<vec.length;i++)
			retVal = add(retVal,mul(vec[i],vec[i]));
		return retVal;
	}
	public void neg(T[] v) {
		for(int i = 0;i<v.length;i++)
			v[i] = neg(v[i]);
	}
	public void decompose(T[] v,T[] b1,T[] b2,T[] par) {
		par[0] = dot(b1,v);
		par[1] = dot(b2,v);
	}
	public void decompose(T[] v,T[][] b,T[] par) {
		for(int i = 0;i<b.length;i++)
			par[i] = dot(b[i],v);
	}
	public void scale(T x,T[] vec) {
		for(int i = 0;i<vec.length;i++)
			vec[i] = mul(x,vec[i]);
	}
	public void invScale(T x,T[] vec) {
		for(int i = 0;i<vec.length;i++)
			vec[i] = div(vec[i],x);
	}
	public void scale(T[] v1,T x,T[] v2) {
		for(int i = 0;i<v1.length;i++)
			v2[i] = mul(x,v1[i]);
	}
	public void scale(T[] scaleVec,T[] v) {
		for(int i = 0;i<scaleVec.length;i++)
			v[i] = mul(v[i],scaleVec[i]);
	}
	public void scale(T[] scaleVec,T[] v1,T[] v2) {
		for(int i = 0;i<scaleVec.length;i++)
			v2[i] = mul(v1[i],scaleVec[i]);
	}
	public void invScale(T[] scaleVec,T[] v) {
		for(int i = 0;i<scaleVec.length;i++)
			v[i] = div(v[i],scaleVec[i]);
	}
	public void add(T[] v1,T[] v2,T[] v3) {
		for(int i = 0;i<v1.length;i++)
			v3[i] = add(v1[i],v2[i]);
	}
	public void mirror(T[] v1,T[] nor,T[] v2) {
		scale(dot(v1,nor),nor);
		scaleAdd(v1,neg(TWO),nor,v2);
	}
	public void rotate(T[] v,boolean dir) {
		if(dir) {
			T tmp = v[v.length-1];
			for(int i = v.length-2;i>=0;i--)
				v[i+1] = v[i];
			v[0] = tmp;
		} else {
			T tmp = v[0];
			for(int i = 0;i<v.length-1;i++)
				v[i] = v[i+1];
			v[v.length-1] = tmp;
		}
	}
	public void rotate2D(T[] vec,T[] ori,T sinphi,T cosphi) {
		T[] tmp = vec.clone();
		diff(vec,ori,tmp);
		vec[0] = add(sub(mul(cosphi,tmp[0]),mul(sinphi,tmp[1])),ori[0]);
		vec[1] = add(add(mul(sinphi,tmp[0]),mul(cosphi,tmp[1])),ori[1]);
	}
	public void rotate2D(T[] vec,T[] ori,T phi) {
		T[] tmp = vec.clone();
		diff(vec,ori,tmp);
		T[] sinCos = sincos(phi);
		vec[0] = add(sub(mul(sinCos[1],tmp[0]),mul(sinCos[0],tmp[1])),ori[0]);
		vec[1] = add(add(mul(sinCos[0],tmp[0]),mul(sinCos[1],tmp[1])),ori[1]);
	}
	public void linBaseComb(T[] par,T[][] vecs,T[] v) {
		for(int i = 0;i<v.length;i++)
			v[i] = ZERO;
		linBaseCombAdd(par,vecs,v);
	}
	public void linBaseCombAdd(T[] par,T[][] vecs,T[] v) {
		for(int i = 0;i<v.length;i++)
			for(int j = 0;j<par.length;j++)
				v[i] = add(v[i],mul(vecs[j][i],par[j]));
	}
	public void linBaseComb(short[] par,T[][] vecs,T[] v) {
		for(int i = 0;i<v.length;i++)
			v[i] = ZERO;
		linBaseCombAdd(par,vecs,v);
	}
	public void linBaseCombAdd(short[] par,T[][] vecs,T[] v) {
		for(int i = 0;i<v.length;i++)
			for(int j = 0;j<par.length;j++)
				v[i] = add(v[i],mul(vecs[j][i],par[j]));
	}
	public void transform(T[] v,T[][] b1,T[][] b2,T[] ret) {
		for(int i = 0;i<ret.length;i++) {
			ret[i] = ZERO;
			for(int j = 0;j<ret.length;j++)
				ret[i] = add(ret[i],mul(dot(b1[j],b2[i]),v[j]));
		}
	}
	public void transform(T[] v,T[][] b,T[] ret) {
		for(int i = 0;i<ret.length;i++) {
			ret[i] = ZERO;
			for(int j = 0;j<ret.length;j++)
				ret[i] = add(ret[i],mul(b[j][i],v[j]));
		}
	}
	public T findRoot(IFunction<T> f,IFunction<T> dfOverDx,T start) throws Exception {
		T cur = util.copy(start,10);
		int curAccu = 0;
		boolean resume;
		do {
			resume = false;
			if(util.precision(cur)<=4*curAccu)
				cur = util.copy(cur,Math.min(4*curAccu,util.precision(start)+10));
			T old = cur;
			cur = approximateRoot(f,dfOverDx,cur);
			T er = abs(sub(cur,old));
			int digits = -util.getScale(er);
			curAccu = Math.max(curAccu,digits);
			resume |= (cmp(er,ZERO)!=0&&digits<=util.precision(start));
		} while(resume);
		return util.copy(cur,util.precision(start));
	}
	public T approximateRoot(IFunction<T> f,IFunction<T> dfOverDx,T cur) throws Exception {
		T delta = div(f.f(cur),dfOverDx.f(cur));
		return sub(cur,delta);
	}
	public T findRoot(IFunction<T> f,T start) throws Exception {
		T cur = util.copy(start,20);
		T old = add(cur,pow(ONE_OVER_TEN,10));
		T fold = f.f(old);
		T subCurOld = sub(cur,old);
		int curAccu = 0;
		boolean resume;
		do {
			resume = false;
			if(util.precision(cur)<=2*curAccu+10)
				cur = util.copy(cur,2*curAccu+10);
			T fcur = f.f(cur);
			T invGrad = div(subCurOld,sub(fcur,fold));
			T val = sub(cur,mul(fcur,invGrad));
			old = cur;
			fold = fcur;
			cur = val;
			subCurOld = sub(cur,old);
			T er = abs(subCurOld);
			int digits = -util.getScale(er);
			curAccu = Math.max(curAccu,digits);
			resume |= (cmp(er,ZERO)!=0&&digits<=util.precision(start));
		} while(resume);
		return util.copy(cur,util.precision(start));
	}
	public float getCalculationSpeed() {
		T pi = util.getPi(5000);
		long t0 = System.currentTimeMillis();
		for(int i = 0;i<66;i++) {
			for(int j = 0;j<3;j++)
				add(pi,pi);
			mul(pi,pi);
		}
		return (System.currentTimeMillis()-t0)*1f/Globals.CALC_TIME_BASE;
	}
}
