/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.PrintStream;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import lattice.Box2D;
import lattice.Circle;
import lattice.IBox;
import lattice.Ray;
import math.Util;
import math.VecMath;
import mvc.scan.Observables;
import mvc.scan.ScanListener;
import pattern.BaseLightPattern;
import pattern.CalculationRecords;
import pattern.LightChar;
import pattern.LightCharPattern;
import pattern.LightPatternSolver.IProgressListener;

import common.ColorTheme;
import common.FileIO;
import common.Globals;
import common.ICallback;
import common.Log;

public final class Controller<T extends Number> {

	public static final byte WIN_YES = 0;
	public static final byte WIN_FULLSCREEN = 1;
	public static final byte WIN_SCREENSAVER = 2;
	public static final byte WIN_MUSEUM = 5;
	public static final byte WIN_MUSEUM_FULL = 6;
	public static final byte WIN_NO_FILE = 3;
	public static final byte WIN_NO_PRINT = 4;
	public static final short UPDATE_DELAY = 200;
	private static final short ANIM_STEPS_PER_SEC = 15;
	
	private final View<T> view;
	private Model<T> model;
	private ScanRunnable scanRun;
	private final CalcReflectionsRunnable calcRun;
	private AnimateRotation<T> animRotRun;
	private AnimatePatterns<T> animPatRun;

	public Controller(VecMath<T> math,
						T radius, 
						BaseLightPattern<T> blp,
						CalculationRecords<T>.Record rec,
						ColorTheme ct,
						byte win) {

		this.calcRun = new CalcReflectionsRunnable();
		this.model = buildModel(math,radius,null);
		initModel(math,blp,rec);
		if(hasWindow(win)){
			ICallback cb = new ICallback() {
				int ctr = 0;
				@Override
				public void call() {
					if(0==ctr++)
						SwingUtilities.invokeLater(new Runnable() {
							public void run() {
								String d = model.getDescription();
								model.setRay(model.getRay(),d,Model.STATIC_CALC);
								startCalculation();
							}
						});
				}
			}; 
			this.view = new View<T>(this,ct,cb,win);
		} else
			view = null;
	}
	public void setPrecision(int precision) {
		Log.getIstc().logln("Precision set to "+precision+". Rebuilding model...");
		VecMath<T> math = model.getMath();
		math.setPrecision(precision);
		List<IModelListener> modelListeners = model.getModelListeners();
		List<ScanListener> scanListeners = model.getScanListeners();
		T ps = model.getShapePercentSize();
		T angle = model.getGridRotation();
		model = buildModel(math,ps,angle);
		for(IModelListener ml:modelListeners) {
			model.addModelListener(ml);
			ml.modelChanged(Model.MODEL_NEW);
		}
		for(ScanListener sl:scanListeners)
			model.addScanListener(sl);
		initModel(math,null,null);
		Log.getIstc().logln("... rebuilding model finished.");
	}
	private Model<T> buildModel(VecMath<T> math,T oldPercentSize,T angle) {
		T[] gridCorner = math.util.array(2);
		gridCorner[0] = math.util.cast(Model.GRID_CORNER_X);
		gridCorner[1] = math.util.cast(Model.GRID_CORNER_Y);
		T[] worldLengths = null;
		if(view!=null){
			Visualizer<T> rv = view.getReflectionVisualizer();
			if(rv!=null&&rv.getPixelMap()!=null){
				int h = rv.getPixelMap().canvasHeight();
				int w = rv.getPixelMap().canvasWidth();
				worldLengths = createNewWorld(w,h,math).getLengths();
			}
		} 
		if(worldLengths==null){
			worldLengths = math.util.array(2);
			worldLengths[0] = math.util.cast(Model.WORLD_SIDE_LENGTH);
			worldLengths[1] = math.util.cast(Model.WORLD_SIDE_LENGTH);
		}
		short points = (short)(Globals.GRID_SIZES[Globals.IDX_GRID_SIZE]+1);
		T size = math.util.copy(math.util.cast(points-1),math.PRECISION);
		T gap = math.div(math.util.cast(Model.GRID_SIDE_LENGTH),size);
		Model<T> m = new Model<T>(gridCorner,points,gap,worldLengths,math);
		m.setShape(new Circle<T>(math.ZERO2D.clone(),math.ONE,math));
		if(Globals.IDX_SHAPE_SIZE!=-1) {
			short shapeSize = Globals.SHAPE_SIZES[Globals.IDX_SHAPE_SIZE];
			m.setShapePercentSize((byte)shapeSize);
		} else {
			T ps = math.util.expand(oldPercentSize,math.PRECISION);
			m.setShapePercentSize(ps);
		}
		if(angle!=null) {
			T phi = math.util.expand(angle,math.PRECISION);
			m.setGridRotation(phi);
		}
		return m;
	}
	Box2D<T> createNewWorld(int w,int h,VecMath<T> mt){
		T msl = mt.util.cast(Model.WORLD_SIDE_LENGTH);
		Box2D<T> newWorld = new Box2D<T>(mt.util.array(2),mt.ONE,mt);
		if(w<=h) {
			newWorld.setLength(msl,(byte)0);
			T newLength = mt.mul(msl,mt.util.cast(h));
			T w_t = mt.util.copy(mt.util.cast(w),mt.PRECISION);
			newLength = mt.div(newLength,w_t);
			newWorld.setLength(newLength,(byte)1);
			T[] corner = mt.util.array(2);
			corner[0] = mt.mul(mt.neg(msl),mt.ONE_OVER_TWO);
			corner[1] = mt.mul(mt.neg(newLength),mt.ONE_OVER_TWO);
			newWorld.setCorner(corner);
		} else {
			newWorld.setLength(msl,(byte)1);
			T newLength = mt.mul(msl,mt.util.cast(w));
			T h_t = mt.util.copy(mt.util.cast(h),mt.PRECISION);
			newLength = mt.div(newLength,h_t);
			newWorld.setLength(newLength,(byte)0);
			T[] corner = mt.util.array(2);
			corner[0] = mt.mul(mt.neg(newLength),mt.ONE_OVER_TWO);
			corner[1] = mt.mul(mt.neg(msl),mt.ONE_OVER_TWO);
			newWorld.setCorner(corner);
		}
		return newWorld;
	}
	private void initModel(VecMath<T> mt,
							BaseLightPattern<T> blp, 
							CalculationRecords<T>.Record rec) {
		T[] start = null;
		if(rec!=null)
			model.setRecord(rec,(short) -1);
		else if(blp != null) {
			IBox<T> gib = model.getGridInitBound();
			T[] corner = gib.getCorner().clone();
			short size = blp.getGridSize();
			T gap = model.calcGap(size);
			T r = model.calcShapeSize(size);
			T maxSize = mt.mul(blp.getMaxShapeSize(),gap);
			maxSize = mt.div(maxSize,mt.util.cast(200));
			if(mt.cmp(r,maxSize)>0)
				Log.getIstc().log("Warning: The max. radius of pattern "+
						blp.getName()+" is lower than the current radius.\n");						
			blp.setParameters(corner,gap,r);
			int curPrec = Globals.PRECISIONS[Globals.IDX_PREC];
			int maxAc = blp.getAccuracyUpperBound();
			if(curPrec<maxAc)
				Log.getIstc().log("Warning: Current precision is to low for "+
						"pattern "+blp.getName()+".\n");
			T phi = mt.util.copy(blp.getRotation(),mt.PRECISION);
			T[] sinCosPhi = model.getMath().sincos(phi);
			model.setGridRotation(phi,sinCosPhi);
			T worldX = model.getWorldBox().getCorner()[0];
			worldX = mt.add(worldX,model.getWorldBox().getLengths()[0]);
			start = blp.getStart(worldX,phi,sinCosPhi);
			start[0] = mt.util.copy(start[0],mt.PRECISION);
			start[1] = mt.util.copy(start[1],mt.PRECISION);
			Ray<T> ray = new Ray<T>(start,model.LIGHT_SOURCE_DIR,mt.util);
			model.setRay(ray,blp.getName(),Model.STATIC_CALC);
		} else{
			Util<T> ut = mt.util;
			start = ut.array(2);
			T worldWidth = model.getWorldBox().getLengths()[0];
			start[0] = ut.round(mt.div(worldWidth,mt.TWO),2);
			start[0] = ut.expand(start[0],mt.PRECISION);
			start[1] = ut.expand(mt.util.parse("-0.214"),mt.PRECISION);
			Ray<T> ray = new Ray<T>(start,model.LIGHT_SOURCE_DIR,mt.util);
			model.setRay(ray,Model.STATIC_CALC);
		}
	}
	public Model<T> getModel() {
		return model;
	}
	public View<T> getView() {
		return view;
	}
	public void startRotAnim(double phiPerMinute) {
		if(animRotRun==null)
			animRotRun = new AnimateRotation<T>(this);
		if(!animRotRun.run)
			synchronized(animRotRun) {
				float dPhi = (float)phiPerMinute/(60*ANIM_STEPS_PER_SEC);
				int pause = 1000/ANIM_STEPS_PER_SEC;
				animRotRun.setPhi(dPhi);
				animRotRun.pause = pause;
				animRotRun.run = true;
				Thread t = new Thread(animRotRun);
				t.setPriority(Thread.MIN_PRIORITY);
				t.start();
			}
	}
	public void stopRotAnim() {
		if(animRotRun!=null)
			animRotRun.run = false;
		synchronized(animRotRun){}
	}
	public void startPatAnim(int startIdx,short gridSize,byte speed) {
		if(animPatRun==null)
			animPatRun = new AnimatePatterns<T>(this,startIdx,gridSize);
		animPatRun.speed = speed;
		if(!animPatRun.run2)
			synchronized(animPatRun) {
				animPatRun.run2 = true;
				Thread t = new Thread(animPatRun);
				t.setPriority(Thread.MIN_PRIORITY);
				t.start();
			}
	}
	public void stopPatAnim() {
		if(animPatRun!=null){
			animPatRun.run2 = false;
			animPatRun.run = false;
		}
		synchronized(animPatRun){
			animPatRun = null;
		}
	}
	public float getRotAnimSpeed(){
		if(animRotRun!=null)
			return animRotRun.getPhi()*60*ANIM_STEPS_PER_SEC;
		return 0;
	}
	public int getPatAnimPos(){
		if(animPatRun!=null)
			return animPatRun.idx;
		return -1;
	}
	public byte getPatAnimSpeed() {
		if(animPatRun!=null)
			return animPatRun.speed;
		return -1;
	}
	public boolean isRotAnimRun() {
		if(animRotRun!=null)
			return animRotRun.run;
		return false;
	}
	public boolean isPatAnimRun() {
		if(animPatRun!=null)
			return animPatRun.run2;
		return false;
	}
	public void setRayLater(Ray<T> r){
		if(isRotAnimRun())
			animRotRun.rayCache.add(r);
	}
	public void startScan(Observables obs,PixelMap<T> pmap) {
		if(scanRun==null)
			scanRun = new ScanRunnable();
		synchronized(scanRun) {
			scanRun.session.stop = false;
			scanRun.session.msg = null;
			scanRun.obs = obs;
			scanRun.pmap = pmap;
			Thread t = new Thread(scanRun);
			t.setPriority(Thread.MIN_PRIORITY);
			t.start();
		}
	}
	public void stopScan(String msg) {
		if(scanRun!=null) {
			scanRun.session.stop = true;
			scanRun.session.msg = msg;
		}
	}
	public boolean stopCalculation() {
		synchronized(model) {
			if(calcRun.requestRun){
				model.setCalculatorActive(false);
				return true;
			}
		}
		return false;
	}
	public void startCalculation() {
		if(calcRun.requestRun==false) {
			// Restrict to single thread execution
			calcRun.requestRun = true;
			Thread t = new Thread(calcRun);
			t.setPriority(Thread.MIN_PRIORITY);
			t.start();
		}
	}
	public void waitForCalculatorRelease() {
		synchronized(calcRun) {
			// wait for a created but not yet executed calculation thread
			while(calcRun.requestRun)
				try {
					calcRun.wait();
				} catch(Exception e) {
					Log.getIstc().log(e);
				}
			return;
		}
	}
	public static <T extends Number> void launch(VecMath<T> mt,String radius,
			String pattern,String theme,String text,String record,byte anim,
			byte win,boolean info){
		FileIO<T> fileio = new FileIO<T>(mt,null);
		if(!hasWindow(win)){
			anim = -1;
			if(record!=null){
				System.out.println("Error: Incompatible options.");
				System.exit(1);
			}
		} else if(win!=WIN_SCREENSAVER){
			String home = System.getProperty("user.home");
			fileio.setUserCfgDir(home+File.separator+".frozenlight");
			install(win,fileio);
		}
		if(anim!=-1){
			pattern = null;
			text = null;
			record = null;
		}
		short size = -1;
		T r = radius!=null?mt.util.parse(radius):null;
		CalculationRecords<T>.Record rec = null;		
		BaseLightPattern<T> blp = null;
		try{
			if(text!=null) {
				LightChar[] lcs = fileio.loadLightChars();
				String s = text.substring(0,Math.min(text.length(),20));
				blp = new LightCharPattern<T>("Text: "+s+"...",text,lcs,mt);
				size = blp.getGridSize();
			} else if(pattern!=null){
				blp = fileio.loadLightPattern(pattern);
				if(blp!=null){
					size = blp.getGridSize();
					if(r==null){
						r = mt.util.parse("50");
						if(mt.cmp(r,blp.getMaxShapeSize())>0)
							r = blp.getMaxShapeSize();
					}
					CalculationRecords<T> recs = fileio.loadRecords(pattern);
					if(recs!=null)
						rec = recs.getRecord(r);
				} else
					throw new Exception("Error: Pattern "+pattern+" not found");
			} else if(record!=null){
				CalculationRecords<T> recs = fileio.loadRecords(new File(record));
				if(recs!=null){
					rec = recs.getRecord(0);
					size = recs.getGridSize();
				} else
					throw new Exception("Error: Record "+record+" not found");
			}
		} catch(Exception e){
			System.out.println(e.getMessage());
			System.exit(1);				
		}
		if(r==null)
			r = mt.util.parse("50");
		Globals.IDX_SHAPE_SIZE = -1;
		if(mt.cmp(r,mt.util.parse(r.byteValue()+""))==0){
			byte idx = Globals.getIndex(r.byteValue(),Globals.SHAPE_SIZES);
			Globals.IDX_SHAPE_SIZE = idx;
		} 
		adjustPrecision(mt,blp,rec,r);
		mt.enableTables();
		if(size!=-1){
			short idx = Globals.getIndex(size,Globals.GRID_SIZES);
			if(idx==-1)
				Globals.addGridSize(size);
			Globals.IDX_GRID_SIZE = 
				Globals.getIndex(size,Globals.GRID_SIZES);
		}
		ColorTheme ct = fileio.loadColorTheme(theme);
		if(ct==null)
			ct = new ColorTheme();
		if(info){
			System.out.println("Calculating pattern "+blp.getName()+"...");
			System.out.println("Estimated time: "+blp.estimateCalcTime()+" sec\n");
			blp.setProgressListener(new ConsoleProgress());
		}
		Controller<T> c = new Controller<T>(mt,r,blp,rec,ct,win);
		if(blp!=null){
			CalculationRecords<T> recs;
			if(win==WIN_NO_FILE){
				recs = c.model.getRecords(blp.getName());
				if(recs==null)
					recs = c.model.createRecord(blp.getName(),mt.PRECISION);
				c.model.addRecords(recs);
				System.exit(0);
			} else if(win==WIN_NO_PRINT){
				recs = c.model.createRecord(blp.getName(),mt.PRECISION);
				List<CalculationRecords<T>.Record> recList = recs.getRecords();
				while(recList.size()>1)
					recList.remove(0);
				recs.setRecords(recList);
				System.out.println(recs.toString());
				System.exit(0);
			}
		}
		MainMenuBar<T> mb = c.getView().getMainMenuBar();	
		if(anim!=-1){
			if(win!=WIN_SCREENSAVER){
				switch(anim){
				case AnimatePatterns.ANIM_PAT_SLOW:
					mb.setSelected(MainMenuBar.ANIM_PAT_ENTRY_SLOW,mb.animMenu,true);
					break;
				case AnimatePatterns.ANIM_PAT_MEDIUM:
					mb.setSelected(MainMenuBar.ANIM_PAT_ENTRY_MEDIUM,mb.animMenu,true);
					break;
				case AnimatePatterns.ANIM_PAT_FAST:
					mb.setSelected(MainMenuBar.ANIM_PAT_ENTRY_FAST,mb.animMenu,true);
					break;
				case AnimatePatterns.ANIM_PAT_VERY_FAST:
					mb.setSelected(MainMenuBar.ANIM_PAT_ENTRY_VERY_FAST,mb.animMenu,true);
					break;
				}
			} else 
				c.startPatAnim(0,(short) -1,anim);
		}	
	}
	static <T extends Number> void adjustPrecision(VecMath<T> mt,BaseLightPattern<T> blp,
			CalculationRecords<T>.Record rec,T r) {
		int prec = -1;
		if(rec!=null){
			T h = mt.util.parse(rec.getHeight().toString());
			prec = mt.util.precision(h);
		} else if(blp!=null){
			T[] corner = mt.util.array(2);
			corner[0] = mt.util.cast(Model.GRID_CORNER_X);
			corner[1] = mt.util.cast(Model.GRID_CORNER_Y);
			short size = blp.getGridSize();
			T s = mt.util.copy(mt.util.cast(size),mt.PRECISION);
			T gap = mt.div(mt.util.cast(Model.GRID_SIDE_LENGTH),s);
			T gapOverTwo = mt.div(gap,mt.TWO);
			T rp = mt.util.copy(mt.div(r,mt.HUNDRED),mt.PRECISION);
			T r0 = mt.mul(gapOverTwo,rp);
			blp.setParameters(corner,gap,r0);
			prec = blp.getAccuracyUpperBound();
		} 
		if(prec!=-1){
			Globals.adjustPrecision(prec);
			mt.setPrecision(Globals.PRECISIONS[Globals.IDX_PREC]);
		} 
	}
	private static <T extends Number> void install(byte win,FileIO<T> fileio) {
		if(!FileIO.checkUserConfigDir()){
			try{
				FileIO.createUserConfigDir();
			} catch(Exception e){
				String txt = "Installation Error. ";
				txt += "Can't create the following directories:\n\n";
				txt += FileIO.getUserCfgDir()+"\n";
				txt += FileIO.getUserPatternDir()+"\n";
				txt += FileIO.getUserRecordDir()+"\n\n";
				txt += "Error message: "+e.getMessage()+"\n";
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				PrintStream ps = new PrintStream(baos);
				e.printStackTrace(ps);
				txt += new String(baos.toByteArray());
				if(hasWindow(win))
					JOptionPane.showMessageDialog(null,txt);
				else
					System.err.println(txt);
				System.exit(1);
			}
			try {
				fileio.installUserConfigs();
			} catch (Exception e) {
				String txt = "Installation Error. Can't copy the ";
				txt += "lightpatterns and records into the directories:\n\n";
				txt += FileIO.getUserPatternDir()+"\n";
				txt += FileIO.getUserRecordDir()+"\n\n";
				txt += "Error message: "+e.getMessage()+"\n";
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				PrintStream ps = new PrintStream(baos);
				e.printStackTrace(ps);
				txt += new String(baos.toByteArray());
				if(hasWindow(win))
					JOptionPane.showMessageDialog(null,txt);
				else
					System.err.println(txt);
			}
		}
	}	
	private static boolean hasWindow(byte win){
		return win==WIN_YES||win==WIN_SCREENSAVER||
				win==WIN_FULLSCREEN||win==WIN_MUSEUM||win==WIN_MUSEUM_FULL;
	}
	static final class ConsoleProgress implements IProgressListener {

		int k = 0;
		public void updateAccuracy(int digits,int maxDigits) {
			int l = (int)(digits*100f/maxDigits);
			if(l>97)
				System.out.print("");
			while(l>k) {
				if(k==0)
					System.out.print("o = 2% : ");
				else { 
					if(k%2==0)
						System.out.print("o");
					if(k%10==0)
						System.out.print(" ");
				}
				if(++k==100)
					System.out.print("o\n");
			}
		}
	}
	final class ScanRunnable implements Runnable {

		ScanSession<T> session = new ScanSession<T>(model);
		PixelMap<T> pmap;
		Observables obs;
		int maxPrecision;
		public synchronized void run() {
			session.init(pmap,obs);
			session.scanObservables();
		}
	}
	final class CalcReflectionsRunnable implements Runnable {

		// Indicates if a thread of this runnable is waiting for
		// execution or is being executed
		private boolean requestRun = false;
		public void run() {
			synchronized(this) {
				try {
					model.calculateReflections();
				} catch(Exception e) {
					Log.getIstc().log(e);
				}
				synchronized(model) {
					model.setCalculatorActive(true);
					requestRun = false;
				}
				notify();
			}
		}
	}
}
