/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JMenu;
import javax.swing.JOptionPane;
import javax.swing.JTextPane;

import math.VecMath;
import pattern.BaseLightPattern;
import pattern.CalculationRecords;
import pattern.ILightPattern;
import pattern.ProgressVisualizer;

import common.Globals;

public final class LightPatternSelector<T extends Number> {

	private final View<T> view;
	private boolean calculating;
	private BaseLightPattern<T> selectedBlp;
	public LightPatternSelector(View<T> v) {
		this.view = v;
	}
	public void select(Component src,BaseLightPattern<T> lp,boolean useRecords) {
		if(!calculating) {
			calculating = true;
			selectedBlp = lp;
			SelectPatternRunnable selPatRun = new SelectPatternRunnable(src,lp,useRecords);
			Thread t = new Thread(selPatRun);
			t.setPriority(Thread.MIN_PRIORITY);
			t.start();
		}
	}
	private class SelectPatternRunnable implements Runnable {

		private final BaseLightPattern<T> pattern;
		private final boolean useRecords;
		private final Component src;
		public SelectPatternRunnable(Component src,BaseLightPattern<T> lp,boolean ur) {
			this.src = src;
			this.pattern = lp;
			this.useRecords = ur;
		}
		public void run() {
			sel(src,pattern,useRecords);
		}
	}
	private void sel(Component src,final BaseLightPattern<T> lp,boolean useRec){
		// Placing this code directly into the mouseReleased method
		// cause problems with painting:
		// If the repaints caused by model.setGridRotation(phi);
		// are waiting to long in the event queue for this mouseReleased
		// to return then the whole paint buffer is dropped away without
		// having a request for background repaint in the reflection vis.
		MainMenuBar<T> menuBar = view.getMainMenuBar();
		menuBar.deselectAnimationButtonGroup(null);
		boolean changeSize = false;
		Controller<T> ctrl = view.getController();
		T percentSize = ctrl.getModel().getShapePercentSize();
		CalculationRecords<T> rec = null;
		int selRec = -1;
		if(useRec) {
			Model<T> model = ctrl.getModel();
			rec = model.getRecords(lp.getName());
			if(rec!=null) {
				StringBuffer msg = new StringBuffer();
				VecMath<T> mt = model.getMath();
				for(int i = 0;i<rec.getRecords().size();i++) {
					T s = rec.getRecords().get(i).getShapeSize();
					if(mt.cmp(mt.abs(mt.sub(percentSize,s)),mt.EPSILON10)<=0) {
						selRec = i;
						break;
					}
					msg.append("    ").append(i).append(")  r = ");
					String num = model.getMath().util.toPlainString(s);
					if(num.length()>16)
						num = num.substring(0,16)+"...";
					msg.append(num).append(" %\n");
				}
				msg.append("\n");
				if(selRec==-1) {
					msg.insert(0,"available for the pattern "+lp.getName()+" :\n\n");
					msg.insert(0,"Stored results of previous calculations are ");
					msg.append("Enter the line number of an already calculated ");
					msg.append("radius r or leave the field empty to make a new ");
					msg.append("calculation for the current radius.");
					JTextPane tp = new JTextPane();
					tp.setText(msg.toString());
					int h = tp.getFontMetrics(tp.getFont()).getHeight();
					Dimension dim = new Dimension(300,130+rec.getRecords().size()*h);
					tp.setPreferredSize(dim);
					tp.setBackground(view.getBackground());
					String line = JOptionPane.showInputDialog(view,tp);
					if(line==null) {
						calculating = false;
						return;
					}
					line = line.replace(")","");
					try {
						selRec = Short.parseShort(line.trim());
						if(0<=selRec&&selRec<rec.getRecords().size())
							changeSize = true;
						else
							selRec = -1;
					} catch(Exception e) {}
				}
			}
		}
		if(selRec!=-1)
			percentSize = rec.getRecords().get(selRec).getShapeSize();
		else if(lp.getMaxShapeSize()!=null) {
			Model<T> model = ctrl.getModel();
			T maxPercentSize = lp.getMaxShapeSize();
			if(model.getMath().cmp(maxPercentSize,percentSize)<0) {
				int choice = JOptionPane.OK_OPTION;
				if(!Globals.AUTO_RADIUS){
					String p1 = model.getMath().util.toPlainString(maxPercentSize);
					T p2Val1 = model.getMath().util.round(percentSize,16);
					T p2Val2 = model.getMath().util.round(percentSize,32);
					String tail = "";
					if(model.getMath().cmp(p2Val1,p2Val2)!=0)
						tail = "...";
					String p2 = model.getMath().util.toPlainString(p2Val1)+tail;
					String msg = "The maximum radius rmax="+p1+"% of the pattern "
					        +lp.getName()+" is smaller than the selected radius"
					        +" r="+p2+"%. Do you wan't to select the maximum"
					        +" pattern radius?";
					JTextPane tp = new JTextPane();
					tp.setText(msg);
					tp.setPreferredSize(new Dimension(300,100));
					tp.setBackground(view.getBackground());
					choice = JOptionPane.showConfirmDialog(view,tp);
				}
				if(choice==JOptionPane.CANCEL_OPTION) {
					calculating = false;
					return;
				}
				if(choice==JOptionPane.OK_OPTION) {
					percentSize = maxPercentSize;
					changeSize = true;
				}
			}
		}
		setCursor(src,Cursor.WAIT_CURSOR);
		setParameters(lp,percentSize);
		int curPrec = Globals.PRECISIONS[Globals.IDX_PREC];
		int lowPrec = Globals.PRECISIONS[Math.max(Globals.IDX_PREC-1,0)];
		int maxAc = lp.getAccuracyUpperBound();
		boolean precHighOk = curPrec>=maxAc; 
		boolean precLowOk = Globals.IDX_PREC==0||lowPrec<=maxAc;
		boolean precOk = curPrec>=maxAc&&lowPrec<=maxAc;
		if(!(precHighOk&&precLowOk)) {
			byte prec_idx = 0;
			while(Globals.PRECISIONS[prec_idx]<maxAc)
				prec_idx++;
			int reqPrec = Globals.PRECISIONS[prec_idx];
			int choice = JOptionPane.OK_OPTION;
			if(!Globals.AUTO_PREC){
				String msg = "The pattern "+lp.getName()+
			        " requires a precision of "+maxAc+" digits. "+
                    "The current precision is "+curPrec+" digits. ";
				if(!precHighOk)
					msg += "Increase precision to "+reqPrec+" digits?";
				else
					msg += "Lower precision to "+reqPrec+" digits?";
				JTextPane tp = new JTextPane();
				tp.setText(msg);
				tp.setPreferredSize(new Dimension(300,90));
				tp.setBackground(view.getBackground());
				choice = JOptionPane.showConfirmDialog(view,tp);
			}
			if(choice==JOptionPane.CANCEL_OPTION) {
				setCursor(src,Cursor.DEFAULT_CURSOR);
				calculating = false;
				return;
			} else if(choice==JOptionPane.OK_OPTION) {
				Globals.IDX_PREC = prec_idx;
				menuBar.setSelectedIdx(Globals.IDX_PREC,menuBar.precMenu,false);
				ctrl.stopCalculation();
				ctrl.waitForCalculatorRelease();
				ctrl.stopScan("Stopped on precision change");
				ctrl.setPrecision(reqPrec);
				if(!changeSize)
					percentSize = ctrl.getModel().getShapePercentSize();
				setParameters(lp,percentSize);
				precOk = true;
			}
		}
		T phi = null;
		if(selRec!=-1)
			phi = rec.getRecords().get(selRec).getAngle();
		else {
			ProgressVisualizer progVis = null;
			if(lp.estimateCalcTime()>=1) {
				progVis = new ProgressVisualizer(lp.getName(),view);
				lp.setProgressListener(progVis);
				progVis.addWindowListener(new WindowAdapter() {

					@Override
					public void windowClosed(WindowEvent e) {
						lp.stopCalc();
					}
					@Override
					public void windowClosing(WindowEvent e) {
						lp.stopCalc();
					}
				});
			}
			phi = lp.getRotation();
			if(progVis!=null)
				progVis.dispose();
		}
		if(phi!=null) {
			Model<T> model = ctrl.getModel();
			VecMath<T> math = model.getMath();
			T phiExt = math.util.copy(phi,math.PRECISION);
			final T[] sinCosPhi = math.sincos(phiExt);
			ctrl.waitForCalculatorRelease();
			if(changeSize) {
				short ps = percentSize.shortValue();
				byte idx = Globals.getIndex(ps,Globals.SHAPE_SIZES);
				if(idx!=-1) {
					short pSize = Globals.SHAPE_SIZES[idx];
					if(math.cmp(percentSize,math.util.cast(pSize))==0) {
						Globals.IDX_SHAPE_SIZE = idx;
						menuBar.setSelectedIdx(idx,menuBar.shapeSizeMenu,false);
					}
				}
				model.setShapePercentSize(percentSize);
			}
			Globals.IDX_GRID_SIZE = Globals.getIndex(lp.getGridSize(),
			        Globals.GRID_SIZES);
			if(Globals.IDX_GRID_SIZE!=-1){
				JMenu gsm = menuBar.gridSizeMenu;
				menuBar.setSelectedIdx(Globals.IDX_GRID_SIZE,gsm,false);
			} else
				menuBar.rebuildGridSizeMenu(lp.getGridSize(),false);
			model.setGridSize(lp.getGridSize());
			model.setGridRotation(phiExt,sinCosPhi);
			T worldX = model.getWorldBox().getCorner()[0];
			worldX = math.add(worldX,model.getWorldBox().getLengths()[0]);
			T[] st = null;
			if(selRec!=-1) {
				st = math.util.array(2);
				st[0] = rec.getRecords().get(selRec).getDistance();
				st[1] = rec.getRecords().get(selRec).getHeight();
			} else {
				st = lp.getStart(worldX,phi,sinCosPhi);
				st[0] = math.util.round(st[0],3);
			}
			st[0] = math.util.copy(st[0],math.PRECISION);
			st[1] = math.util.copy(st[1],math.PRECISION);
			model.getRay().setStart(st);
			model.setRay(model.getRay(),lp.getName(),Model.STATIC_CALC);
			ctrl.startCalculation();
			if(useRec&&selRec==-1&&precOk) {
				if(rec==null)
					rec = new CalculationRecords<T>(lp.getName(),
													lp.getGridSize(),
													math);
				T h = math.util.copy(st[1],maxAc);
				T angle = math.util.copy(phi,maxAc);
				rec.add(percentSize,math.util.round(st[0],4),h,angle);
				model.addRecords(rec);
			}
		}
		setCursor(src,Cursor.DEFAULT_CURSOR);
		calculating = false;		
	}
	private void setCursor(Component src,int cursor){
		if(src!=null)
			src.setCursor(Cursor.getPredefinedCursor(cursor));
		view.setCursor(Cursor.getPredefinedCursor(cursor));		
	}
	private void setParameters(ILightPattern<T> pattern,T percentSize) {
		Model<T> model = view.getController().getModel();
		VecMath<T> mt = model.getMath();
		T gap = model.calcGap(pattern.getGridSize());
		T size = mt.mul(percentSize,gap);
		size = mt.div(size,mt.util.cast(200));
		T[] cor = model.getGridInitBound().getCorner().clone();
		pattern.setParameters(cor,gap,size);
	}
	public BaseLightPattern<T> getSelected() {
		return selectedBlp;
	}
}
