/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package mvc.scan;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.font.FontRenderContext;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.HashMap;

import javax.swing.JComponent;
import javax.swing.border.MatteBorder;

import lattice.Box2D;
import math.VecMath;
import mvc.Controller;
import mvc.IModelListener;
import mvc.Model;
import mvc.PixelMap;
import mvc.var.VariableVisualizer;

public class ScanVisualizer<T extends Number> extends JComponent {

	private final short YSIDE_PIXEL = 300;
	private final short XSIDE_PIXEL = 300;
	private final byte BORDER = 10;
	private final Controller<T> ctrl;
	private final Box2D<T> iniPhiYBox;
	private final Box2D<T> curPhiYBox;
	private final HashMap<Short,Color> colors;
	private PixelMap<T> trafo;
	private Box2D<T> dragBox;
	private Observables obs;
	private Observables unscaledObs;
	private byte observable;
	private ScanListener scanListener;
	private IModelListener modelListener;
	private BufferedImage bimg;
	private String message;
	public ScanVisualizer(Controller<T> c) {
		this.ctrl = c;
		this.observable = Observables.OBS_REFL_CNT;
		Model<T> model = ctrl.getModel();
		VecMath<T> math = model.getMath();
		T[] corner = math.util.array(2);
		corner[0] = math.neg(math.PI_OVER_FOUR);
		corner[1] = model.getWorldBox().getCorner()[1];
		T[] lengths = math.util.array(2);
		lengths[0] = math.PI_OVER_TWO;
		lengths[1] = model.getWorldBox().getLengths()[1];
		this.iniPhiYBox = new Box2D<T>(corner,lengths,math);
		this.curPhiYBox = iniPhiYBox;
		this.colors = new HashMap<Short,Color>();
		setBorder(new MatteBorder(BORDER,BORDER,BORDER,BORDER,Color.black));
		setOpaque(true);
		initListeners();
	}
	private void initListeners() {
		ZoomMouseListener zl = new ZoomMouseListener();
		addMouseMotionListener(zl);
		addMouseListener(zl);
		modelListener = new IModelListener() {

			@Override
			public void modelChanged(byte event) {
				switch(event) {
					case Model.LIGHTPATTERNS_UPDATE:
						break;
					case Model.MODEL_WORLD_RESIZE:
						// trafo.worldChanged();
						// TODO Scanner shouldn't depend on resizing?
					case Model.MODEL_NEW:
						trafo = null;
						break;
					case Model.GRID_RESIZE:
						stopScan("Stopped on grid change");
						break;
					case Model.SHAPE_RESIZE:
						stopScan("Stopped on circle change");
						break;
				}
			}
		};
		ctrl.getModel().addModelListener(modelListener);
		scanListener = new ScanListener() {

			short width;
			short height;
			@Override
			public void start() {
				message = null;
				height = getPixelMap().canvasHeight();
				width = getPixelMap().canvasWidth();
				// newObservableImage();
				repaint();
			}
			@Override
			public void stop(String msg) {
				if(msg!=null) {
					message = msg;
					repaint();
				}
			}
			@Override
			public void update(int x,int y) {
				if(x<=width-1) {
					if(obs.isMaxChanged(observable)) {
						obs.setMaxChanged(observable,false);
						paintObservableImage(0,0,x+1,bimg.getHeight());
						repaint(BORDER,BORDER,x+1,height);
					} else {
						paintObservableImage(x,y,x+1,y+1);
						repaint(BORDER+x,BORDER+y,1,1);
					}
				}
			}
		};
		ctrl.getModel().addScanListener(scanListener);
		addMouseListener(new SelectMouseListener());
	}
	public void paintComponent(Graphics g) {
		Graphics2D g2d = (Graphics2D)g;
		PixelMap<T> pixMap = getPixelMap();
		pixMap.setSreenSize((short)getWidth(),(short)getHeight());
		final short h = pixMap.canvasHeight();
		final short w = pixMap.canvasWidth();
		if(obs==null) {
			obs = new Observables(w,h,obs);
			unscaledObs = obs;
			newObservableImage();
		}
		if(obs.reflCnt.length!=w||obs.reflCnt[0].length!=h) {
			stopScan("Stopped on resize");
			obs = new Observables(w,h,obs);
			obs.scaledCopy(this.unscaledObs);
			newObservableImage();
		}
		g2d.drawImage(bimg,BORDER,BORDER,this);
		if(dragBox!=null) {
			g2d.setColor(Color.black);
			Rectangle r = pixMap.box2Pix(dragBox);
			g2d.drawRect(r.x,r.y,r.width,r.height);
		}
		String msg = message;
		if(msg!=null) {
			g2d.setColor(Color.BLACK);
			FontRenderContext rctx = g2d.getFontRenderContext();
			Rectangle2D bnd = g2d.getFont().getStringBounds(msg,rctx);
			int textWidth = getWidth()-(int)bnd.getWidth()-BORDER-2;
			int textHeight = getHeight()-BORDER-2;
			g2d.drawString(msg,textWidth,textHeight);
		}
	}
	private void paintObservableImage(int xMin,int yMin,int xMax,int yMax) {
		int w = bimg.getWidth();
		int h = bimg.getHeight();
		if(xMin<w&&xMax<=w&&yMin<h&&yMax<=h)
			for(int x = xMin;x<xMax;x++)
				for(int y = yMin;y<yMax;y++) {
					Color color = null;
					switch(observable) {
						case Observables.OBS_REFL_CNT:
							color = getColor(obs.reflCnt[x][y],obs.maxReflCnt);
							break;
						case Observables.OBS_LENGTH:
							color = getColor(obs.length[x][y],obs.maxLength);
							break;
						case Observables.OBS_OBS1:
							color = getColor(obs.obs1[x][y],obs.maxObs1);
					}
					bimg.setRGB(x,y,color.getRGB());
				}
	}
	private void newObservableImage() {
		PixelMap<T> pixMap = getPixelMap();
		final short h = pixMap.canvasHeight();
		final short w = pixMap.canvasWidth();
		bimg = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
		paintObservableImage(0,0,bimg.getWidth(),bimg.getHeight());
	}
	public void prepareScan(T[] p1,T[] p2) {
		setBoxCorners(p1,p2,curPhiYBox);
		getPixelMap().worldChanged(curPhiYBox);
	}
	private void setBoxCorners(T[] start,T[] end,Box2D<T> box) {
		Model<T> model = ctrl.getModel();
		T[] corner = model.getMath().util.array(2);
		VecMath<T> math = model.getMath();
		corner[0] = math.cmp(start[0],end[0])<0?start[0]:end[0];
		corner[1] = math.cmp(start[1],end[1])<0?start[1]:end[1];
		T[] lengths = model.getMath().util.array(2);
		lengths[0] = math.abs(math.sub(start[0],end[0]));
		lengths[1] = math.abs(math.sub(start[1],end[1]));
		box.setCorner(corner);
		box.setLengths(lengths);
	}
	public void startScan() {
		final short h = getPixelMap().canvasHeight();
		final short w = getPixelMap().canvasWidth();
		startScan(w,h);
	}
	private void startScan(final int width,final int height) {
		this.obs = new Observables(width,height,obs);
		unscaledObs = obs;
		newObservableImage();
		ctrl.startScan(obs,getPixelMap());
	}
	public void stopScan(String msg) {
		ctrl.stopScan(msg);
	}
	private Color getColor(float val,float maxVal) {
		// Debug ...
		if(val>maxVal)
			return Color.red; //...
		short relVal = (short)(val/maxVal*1000);
		Color c = colors.get(relVal);
		if(c==null) {
			float x = (float)(1-Math.sqrt(relVal/1000f));
			Color sc = Color.white;
			c = new Color((int)(sc.getRed()*x),(int)(sc.getGreen()*x),(int)(sc.getBlue()*x));
			colors.put(relVal,c);
		}
		return c;
	}
	public PixelMap<T> getPixelMap() {
		if(trafo==null) {
			Model<T> model = ctrl.getModel();
			VecMath<T> mt = model.getMath();
			trafo = new PixelMap<T>(getWidth(),getHeight(),BORDER,BORDER,curPhiYBox,mt);
		}
		return trafo;
	}
	@Override
	public Dimension getPreferredSize() {
		return new Dimension(XSIDE_PIXEL+BORDER,YSIDE_PIXEL+BORDER);
	}
	public void setObservable(byte o) {
		this.observable = o;
		VariableVisualizer<T> varVis = ctrl.getView().getVariableVisualizer();
		if(varVis!=null)
			varVis.setObservable(o,(obs!=null)?obs.getMax(o):0);
		if(obs!=null) {
			obs.setMaxChanged(observable,false);
			if(bimg!=null) {
				paintObservableImage(0,0,bimg.getWidth(),bimg.getHeight());
				repaint();
			}
		}
	}
	public byte getObservable() {
		return observable;
	}
	public Observables getObservables() {
		return obs;
	}
	public void onDelete() {
		ctrl.getModel().removeScanListener(scanListener);
		ctrl.getModel().removeModelListener(modelListener);
	}
	public void addScanListener(ScanListener l) {
		ctrl.getModel().addScanListener(l);
	}
	public void removeScanListener(ScanListener l) {
		ctrl.getModel().removeScanListener(l);
	}
	public class SelectMouseListener extends MouseAdapter {

		@Override
		public void mouseClicked(MouseEvent e) {
			Model<T> model = ctrl.getModel();
			ctrl.stopCalculation();
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			PixelMap<T> pixMap = getPixelMap();
			T phi = pixMap.xPix2World(pixMap.toCanvasX((short)e.getX()));
			VecMath<T> mt = model.getMath();
			T[] sincosphi = mt.sincos(phi);
			T[] startVec = model.getRay().getStart();
			Box2D<T> world = model.getWorldBox();
			startVec[0] = mt.add(world.getCorner()[0],world.getLengths()[0]);
			startVec[1] = pixMap.yPix2World(pixMap.toCanvasY((short)e.getY()));
			ctrl.waitForCalculatorRelease();
			model.setGridRotation(phi,sincosphi);
			ctrl.startCalculation();
			setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		}
	}
	public class ZoomMouseListener extends MouseAdapter {

		Point start;
		@Override
		public void mouseDragged(MouseEvent e) {
			if(start==null) {
				start = e.getPoint();
			} else {
				VecMath<T> math = ctrl.getModel().getMath();
				Point end = e.getPoint();
				Rectangle r = dragBox!=null?getPixelMap().box2Pix(dragBox):null;
				dragBox = null;
				if(Math.abs(start.x-end.x)>10&&Math.abs(start.y-end.y)>10) {
					dragBox = new Box2D<T>(math.util.array(2),math.ONE,math);
					setBoxCorners(toWorld(start),toWorld(end),dragBox);
					Rectangle r2 = getPixelMap().box2Pix(dragBox);
					r = r!=null?r.union(r2):r2;
				}
				if(r!=null)
					repaint(r.x,r.y,r.width+1,r.height+1);
			}
		}
		@Override
		public void mouseReleased(MouseEvent e) {
			if(start!=null&&dragBox!=null) {
				stopScan("Stopped on zoom");
				prepareScan(toWorld(start),toWorld(e.getPoint()));
				PixelMap<T> pixMap = getPixelMap();
				startScan(pixMap.canvasWidth(),pixMap.canvasHeight());
			}
			dragBox = null;
			start = null;
		}
		private T[] toWorld(Point p) {
			Model<T> model = ctrl.getModel();
			T[] q = model.getMath().util.array(2);
			PixelMap<T> pm = getPixelMap();
			q[0] = pm.xPix2World(pm.toCanvasX((short)p.x));
			q[1] = pm.yPix2World(pm.toCanvasY((short)p.y));
			return q;
		}
	}
}
