/**
	Copyright (c) 2009-2013 by Zoltan Palmer

	This file is part of FroZenLight.

    FroZenLight is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FroZenLight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FroZenLight.  If not, see http://www.gnu.org/licenses/
*/
package pattern;
import java.util.ArrayList;
import java.util.List;

import lattice.Circle;
import lattice.Ray;
import lattice.RayPolygon;
import math.VecMath;
import pattern.LightPatternSolver.IProgressListener;

import common.Globals;
import common.Log;

public class BaseLightPattern<T extends Number> extends LightPatternSequence
        implements
            ILightPattern<T> {

	public final static byte POINT = 0;
	public final static byte GRAD_RIGHT = 1;
	public final static byte GRAD_LEFT = 2;
	private boolean log;
	private T startBndVal;
	private T endBndVal;
	private T[] endPt;
	private T[] startPt;
	private T[] reflPt;
	private byte startBndValType;
	private byte endBndValType;
	private T[] reflPt2;
	private String name;
	private T maxShapeSize;
	private T m;
	private IProgressListener pl;
	protected LightPatternSolver<T> solver;
	protected final VecMath<T> mt;
	protected short gridSize;
	protected T gap;
	protected T r;
	protected T[] corner;
	protected int acUpperBnd;
	private List<T[]> reflPts;
	private int acLowerBnd;
	private boolean storeCalc;
	private int startIndex;
	private boolean propagating;
	public BaseLightPattern(VecMath<T> math) {
		this.log = true;
		this.mt = math;
		this.name = "";
		this.startBndVal = math.ZERO;
		this.endBndVal = math.ZERO;
		this.startBndValType = POINT;
		this.endBndValType = POINT;
		this.gridSize = 0;
		this.maxShapeSize = math.ZERO;
		this.storeCalc = false;
		this.startIndex = -1;
		this.propagating = false;
	}
	public BaseLightPattern(String name,VecMath<T> math) {
		this(math);
		this.name = name;
	}
	public BaseLightPattern(BaseLightPattern<T> p) {
		super(p);
		this.log = true;
		this.name = p.name;
		this.mt = p.mt;
		this.startBndVal = p.startBndVal;
		this.endBndVal = p.endBndVal;
		this.startBndValType = p.startBndValType;
		this.endBndValType = p.endBndValType;
		this.gridSize = p.gridSize;
		this.maxShapeSize = p.maxShapeSize;
		this.storeCalc = p.storeCalc;
		this.startIndex = p.startIndex;
		this.propagating = p.propagating;
	}
	public boolean equals(BaseLightPattern<T> p) {
		boolean b = super.equals(p);
		b &= mt.cmp(startBndVal,p.startBndVal)==0;
		b &= mt.cmp(endBndVal,p.endBndVal)==0;
		b &= startBndValType==p.startBndValType;
		b &= endBndValType==p.endBndValType;
		b &= gridSize==p.gridSize;
		b &= mt.cmp(maxShapeSize,p.maxShapeSize)==0;
		b &= storeCalc==p.storeCalc;
		b &= propagating==p.propagating;
		return b;
	}
	@Override
	public void setParameters(T[] gridCorner,T g,T rs) {
		boolean precChange = false;
		if(gap!=null&&r!=null) {
			precChange = g.doubleValue()==gap.doubleValue();
			precChange &= rs.doubleValue()==r.doubleValue();
		}
		this.corner = gridCorner;
		this.gap = g;
		this.r = rs;
		solver = new LightPatternSolver<T>(r.floatValue()/gap.floatValue(),false,mt);
		if(!precChange) {
			boolean oldLog = log;
			log = false;
			int[] lossLowHigh = calcDigitLossLowHigh();
			log = oldLog;
			// Observation 1: circles 253,spread 13
			// Observation 2: circles 815,spread 50
			int spread = (int)Math.sqrt(2*cells.size());
			this.acLowerBnd = lossLowHigh[0]+Globals.CALC_ACCURACY;
			this.acUpperBnd = lossLowHigh[1]+Globals.CALC_ACCURACY+spread;
			if(log)
				Log.getIstc().logln("Pattern "+name+" : required accuracy = 1e-"+acUpperBnd);
		}
		int ac = isPropagating()?Math.min(acUpperBnd,mt.PRECISION-5):mt.PRECISION-5;
		solver.setAccuracy(ac);
		solver.setLog(log);
		solver.addProgressListener(pl);
	}
	@Override
	public int estimateCalcTime() {
		float speed = mt.getCalculationSpeed();
		speed /= (Runtime.getRuntime().availableProcessors()*0.8f);
		double t = 10.88*Math.exp(-2.63*solver.getOptimalDamp().floatValue());
		t = t*(2.7e-6*acUpperBnd*acUpperBnd+7.1e-4*acUpperBnd+0.13);
		t = t/(2.7e-6*500*500+7.1e-4*500+0.13);
		t = speed*m_init.size()*t;
		int o = (int)Math.log10(t);
		if(o>=2)
			t = Math.pow(10,o-1)*(int)(t/Math.pow(10,o-1)+0.5);
		return (int)(t+0.5);
	}
	public void setProgressListener(IProgressListener l) {
		pl = l;
		if(solver!=null)
			solver.addProgressListener(pl);
	}
	@Override
	public T[] getStart(T worldMaxX,T phi,T[] sinCosPhi) {
		T[] retVal = null;
		if(getMaxShapeSize()==null||mt.cmp(r,getMaxShapeSize())<=0) {
			retVal = reflPt.clone();
			mt.rotate2D(retVal,mt.ZERO2D,sinCosPhi[0],sinCosPhi[1]);
			if(isPropagating())
				retVal[0] = worldMaxX;
			else {
				T gapDiv2 = mt.div(gap,mt.TWO);
				retVal[0] = mt.add(retVal[0],mt.sub(gapDiv2,r));
			}
		}
		return retVal;
	}
	@Override
	public boolean isPropagating() {
		return propagating;
	}
	public void setPropagating(boolean b) {
		this.propagating = b;
	}
	@Override
	public T getRotation() {
		T retVal = null;
		if(getMaxShapeSize()==null||mt.cmp(r,getMaxShapeSize())<=0) {
			if(startBndValType==POINT&&endBndValType==POINT)
				calcGradientPP();
			else if(startBndValType==POINT&&isGradBndType(endBndValType))
				calcGradientPG();
			else if(isGradBndType(startBndValType)&&endBndValType==POINT)
				calcGradientGP();
			else if(isGradBndType(startBndValType)&&isGradBndType(endBndValType))
				calcGradientGG();
			if(m!=null) {
				T phi = mt.atan(m);
				T dy = null;
				T dx = null;
				if(reflPt2!=null) {
					dy = mt.sub(reflPt[1],reflPt2[1]);
					dx = mt.sub(reflPt[0],reflPt2[0]);
				} else {
					boolean dir;
					if(getStartIndex()==0) {
						dir = startBndValType==GRAD_RIGHT;
						dy = dir?startBndVal:mt.neg(startBndVal);
					} else {
						dir = endBndValType==GRAD_LEFT;
						dy = dir?endBndVal:mt.neg(endBndVal);
					}
					dx = dir?mt.ONE:mt.MINUS_ONE;
				}
				if(mt.cmp(dy,mt.ZERO)==0)
					retVal = mt.cmp(dx,mt.ZERO)>0?mt.PI:mt.ZERO;
				else {
					boolean ydir = mt.cmp(dy,mt.ZERO)>0;
					if(mt.cmp(phi,mt.ZERO)>0)
						retVal = ydir?mt.sub(mt.PI,phi):mt.neg(phi);
					else
						retVal = ydir?mt.neg(phi):mt.sub(mt.neg(phi),mt.PI);
				}
			}
		}
		return retVal;
	}
	private static boolean isGradBndType(byte bt) {
		return bt==GRAD_LEFT||bt==GRAD_RIGHT;
	}
	private List<Circle<T>> buildCircles(List<short[]> cellList,int i0,int i1) {
		ArrayList<Circle<T>> circles = new ArrayList<Circle<T>>();
		T gapOver2 = mt.div(gap,mt.TWO);
		for(int i = i0;i<i1;i++) {
			T[] center = mt.util.array(2);
			center[0] = mt.add(mt.mul(gap,cellList.get(i)[0]),gapOver2);
			center[1] = mt.add(mt.mul(gap,cellList.get(i)[1]),gapOver2);
			mt.add(center,corner,center);
			Circle<T> circle = new Circle<T>(center,r,mt);
			circles.add(circle);
		}
		return circles;
	}
	private void calcGradientPP() {
		startPt = calcBoundPoint(cells.get(0),startBndVal.byteValue());
		endPt = calcBoundPoint(cells.get(cells.size()-1),endBndVal.byteValue());
		if(cells.size()>2) {
			List<Circle<T>> circles = buildCircles(cells,1,cells.size()-1);
			reflPts = solver.solve(startPt,endPt,circles,m_init);
			if(solver.getTerminated()) {
				m = null;
				return;
			}
			int ac;
			int idx = getStartIndex();
			if(idx<=1) {
				reflPt = reflPts.get(0);
				reflPt2 = startPt;
				ac = solver.getNodeAccuracy(0);
			} else if(idx==cells.size()-1) {
				reflPt = endPt;
				reflPt2 = reflPts.get(reflPts.size()-1);
				ac = solver.getNodeAccuracy(reflPts.size()-1);
			} else {
				reflPt = reflPts.get(idx-1);
				reflPt2 = reflPts.get(idx-2);
				ac = Math.min(solver.getNodeAccuracy(idx-1),solver.getNodeAccuracy(idx-2));
			}
			if(log)
				Log.getIstc().logln("Accuracy Spread = "+(solver.getAccuracy()-ac));
		} else {
			reflPts = new ArrayList<T[]>();
			reflPt = endPt;
			reflPt2 = startPt;
		}
		m = mt.div(mt.sub(reflPt[1],reflPt2[1]),mt.sub(reflPt[0],reflPt2[0]));
	}
	private void calcGradientPG() {
		startPt = calcBoundPoint(cells.get(0),startBndVal.byteValue());
		boolean boundDir = endBndValType==GRAD_RIGHT;
		T val = endBndVal;
		if(cells.size()>1) {
			List<Circle<T>> circles = buildCircles(cells,1,cells.size());
			reflPts = solver.solve(startPt,val,boundDir,circles,m_init);
			if(solver.getTerminated()) {
				m = null;
				return;
			}
			int ac;
			int idx = getStartIndex();
			if(idx<=1) {
				reflPt = reflPts.get(0);
				reflPt2 = startPt;
				ac = solver.getNodeAccuracy(0);
			} else {
				reflPt = reflPts.get(idx-1);
				reflPt2 = reflPts.get(idx-2);
				ac = Math.min(solver.getNodeAccuracy(idx-1),solver.getNodeAccuracy(idx-2));
			}
			if(log)
				Log.getIstc().logln("Accuracy Spread = "+(solver.getAccuracy()-ac));
			val = mt.div(mt.sub(reflPt[1],reflPt2[1]),mt.sub(reflPt[0],reflPt2[0]));
		} else
			reflPt = startPt;
		m = mt.util.truncate(val,solver.getAccuracy());
	}
	private void calcGradientGP() {
		endPt = calcBoundPoint(cells.get(cells.size()-1),endBndVal.byteValue());
		boolean boundDir = startBndValType==GRAD_RIGHT;
		T val = startBndVal;
		if(cells.size()>1) {
			List<Circle<T>> circles = buildCircles(cells,0,cells.size()-1);
			reflPts = solver.solve(val,boundDir,endPt,circles,m_init);
			if(solver.getTerminated()) {
				m = null;
				return;
			}
			int ac;
			int idx = getStartIndex();
			if(idx==0) {
				reflPt = reflPts.get(0);
				ac = solver.getNodeAccuracy(0);
			} else {
				reflPt = reflPts.get(idx);
				reflPt2 = reflPts.get(idx-1);
				val = mt.div(mt.sub(reflPt[1],reflPt2[1]),mt.sub(reflPt[0],reflPt2[0]));
				ac = Math.min(solver.getNodeAccuracy(idx),solver.getNodeAccuracy(idx-1));
			}
			if(log)
				Log.getIstc().logln("Accuracy Spread = "+(solver.getAccuracy()-ac));
		} else
			reflPt = endPt;
		m = mt.util.truncate(val,solver.getAccuracy());
	}
	private void calcGradientGG() {
		boolean boundDir1 = startBndValType==GRAD_RIGHT;
		boolean boundDir2 = endBndValType==GRAD_RIGHT;
		boolean[] boundDirs = new boolean[]{boundDir1,boundDir2};
		T[] boundValues = mt.util.array(2);
		boundValues[0] = startBndVal;
		boundValues[1] = endBndVal;
		T val = boundValues[0];
		if(cells.size()>0) {
			List<Circle<T>> circles = buildCircles(cells,0,cells.size());
			reflPts = solver.solve(boundValues,boundDirs,circles,m_init);
			if(solver.getTerminated()) {
				m = null;
				return;
			}
			int ac;
			int idx = getStartIndex();
			if(idx==0) {
				reflPt = reflPts.get(0);
				ac = solver.getNodeAccuracy(0);
			} else if(idx==cells.size()) {
				reflPt = reflPts.get(cells.size()-1);
				ac = solver.getNodeAccuracy(cells.size()-1);
				val = boundValues[1];
			} else {
				reflPt = reflPts.get(idx);
				reflPt2 = reflPts.get(idx-1);
				val = mt.div(mt.sub(reflPt[1],reflPt2[1]),mt.sub(reflPt[0],reflPt2[0]));
				ac = Math.min(solver.getNodeAccuracy(idx),solver.getNodeAccuracy(idx-1));
			}
			if(log)
				Log.getIstc().logln("Accuracy Spread = "+(solver.getAccuracy()-ac));
		}
		m = mt.util.truncate(val,solver.getAccuracy());
	}
	private T[] calcBoundPoint(short[] cell,byte bndIdx) {
		T[] retVal = mt.util.array(2);
		T gapOver2 = mt.div(gap,mt.TWO);
		retVal[0] = mt.add(mt.mul(gap,cell[0]),gapOver2);
		retVal[1] = mt.add(mt.mul(gap,cell[1]),gapOver2);
		mt.add(retVal,corner,retVal);
		T dxy = mt.div(r,mt.sqrt(mt.TWO,mt.PRECISION));
		switch(bndIdx) {
			case 0:
				retVal[0] = mt.add(retVal[0],r);
				break;
			case 1:
				retVal[0] = mt.add(retVal[0],dxy);
				retVal[1] = mt.add(retVal[1],dxy);
				break;
			case 2:
				retVal[1] = mt.add(retVal[1],r);
				break;
			case 3:
				retVal[0] = mt.sub(retVal[0],dxy);
				retVal[1] = mt.add(retVal[1],dxy);
				break;
			case 4:
				retVal[0] = mt.sub(retVal[0],r);
				break;
			case 5:
				retVal[0] = mt.sub(retVal[0],dxy);
				retVal[1] = mt.sub(retVal[1],dxy);
				break;
			case 6:
				retVal[1] = mt.sub(retVal[1],r);
				break;
			case 7:
				retVal[0] = mt.add(retVal[0],dxy);
				retVal[1] = mt.sub(retVal[1],dxy);
				break;
		}
		return retVal;
	}
	public void setStartIndex(int idx) {
		this.startIndex = idx;
	}
	public void setStartBndVal(T sbv) {
		this.startBndVal = sbv;
	}
	public void setEndBndVal(T ebv) {
		this.endBndVal = ebv;
	}
	public void setStartBndValType(byte bt) {
		this.startBndValType = bt;
	}
	public void setEndBndValType(byte bt) {
		this.endBndValType = bt;
	}
	public void setGridSize(short s) {
		this.gridSize = s;
	}
	public void setName(String name) {
		this.name = name;
	}
	public void setMaxShapeSize(T maxSize) {
		this.maxShapeSize = maxSize;
	}
	public void setStoreCalculation(boolean b) {
		this.storeCalc = b;
	}
	@Override
	public int getStartIndex() {
		return startIndex;
	}
	@Override
	public T getStartBndVal() {
		return startBndVal;
	}
	@Override
	public T getEndBndVal() {
		return endBndVal;
	}
	@Override
	public byte getStartBndValType() {
		return startBndValType;
	}
	@Override
	public byte getEndBndValType() {
		return endBndValType;
	}
	@Override
	public short getGridSize() {
		return gridSize;
	}
	@Override
	public String getName() {
		return name;
	}
	@Override
	public T getMaxShapeSize() {
		return maxShapeSize;
	}
	@Override
	public boolean isStoreCalcEnabled() {
		return storeCalc;
	}
	public boolean check() {
		String msg = null;
		if(cells==null)
			msg = "ERROR: No cells have been defined for pattern "+getName();
		else if(startBndValType==POINT&&endBndValType==POINT) {
			if(cells.size()<2)
				msg = "ERROR: Pattern "+getName()+" must have at least 2 cells";
			else if(cells.size()<m_init.size()+2)
				msg = "ERROR: Found to many initial gradients in pattern "+getName();
			else if(cells.size()>m_init.size()+2)
				msg = "ERROR: Missing initial gradients in pattern "+getName();
		} else if(!(cells.size()>=1))
			msg = "ERROR: Pattern "+getName()+" must have at least 1 cell";
		if(msg!=null) {
			Log.getIstc().logln(msg);
			return false;
		}
		return true;
	}
	public void stopCalc() {
		if(solver!=null)
			solver.setTerminated(true);
	}
	public void setAccuracy(int ac) {
		solver.setAccuracy(ac);
	}
	@Override
	public int getAccuracy() {
		return solver.getAccuracy();
	}
	@Override
	public int getAccuracyUpperBound() {
		return acUpperBnd;
	}
	@Override
	public int getAccuracyLowerBound() {
		return acLowerBnd;
	}
	protected int[] calcDigitLossLowHigh() {
		float[] lossLowHigh = calcSequenceDigitLossLowHigh();
		float dll = lossLowHigh[0];
		if(!isPropagating())
			dll *= getRotations();
		// The errors which appear at the bonds of the different
		// rotations are neglected here.
		float dlh = lossLowHigh[1];
		if(isPropagating()) {
			// Worst case error estimation
			dlh *= gridSize;
			if(startBndValType!=POINT||endBndValType!=POINT)
				dlh *= 2;
		} else
			dlh *= getRotations();
		int[] errors = new int[]{(int)dll,(int)(dlh+0.5)};
		return errors;
	}
	protected float[] calcSequenceDigitLossLowHigh() {
		solver.setAccuracy(LightPatternSolver.INI_AC/4);
		getRotation();
		RayPolygon<T> rp = new RayPolygon<T>(mt.util,solver.getAccuracy());
		rp.setShape(new Circle<T>(mt.ZERO2D.clone(),r,mt));
		if(startBndValType==POINT) {
			Circle<T> c = buildCircles(cells,0,1).get(0);
			if(reflPts.size()>0)
				rp.add(mirrorPoint(c.getCenter(),startPt,reflPts.get(0)));
			else if(endBndValType==POINT)
				rp.add(mirrorPoint(c.getCenter(),startPt,endPt));
			rp.add(startPt);
		} else
			rp.setInRay(boundRay(reflPts.get(0),startBndVal));
		for(int i = 0;i<reflPts.size();i++)
			rp.add(reflPts.get(i));
		if(endBndValType==POINT) {
			rp.add(endPt);
			if(reflPts.size()>0) {
				Circle<T> c = buildCircles(cells,cells.size()-1,cells.size()).get(0);
				rp.add(mirrorPoint(c.getCenter(),endPt,reflPts.get(reflPts.size()-1)));
			}
		} else
			rp.setOutRay(boundRay(reflPts.get(reflPts.size()-1),endBndVal));
		float dlh = rp.getDigitLossHigh();
		float dllTmp = rp.getDigitLossLow();
		if(startBndValType==POINT)
			rp.remove(0);
		if(endBndValType==POINT)
			rp.removeLast();
		if(!isPropagating())
			dlh -= (dlh-rp.getDigitLossHigh())/2;
		float dll = rp.getDigitLossLow();
		dll += (dllTmp-dll)/2;
		return new float[]{dll,dlh};
	}
	private Ray<T> boundRay(T[] a,T grad) {
		T[] v = mt.util.array(2);
		v[0] = mt.ONE;
		v[1] = grad;
		mt.normalize(v);
		return new Ray<T>(a,v,mt.util);
	}
	private T[] mirrorPoint(T[] ctr,T[] bndPt,T[] prevPt) {
		T[] n = mt.util.array(2);
		mt.diff(bndPt,ctr,n);
		mt.normalize(n);
		T[] x = mt.util.array(2);
		mt.diff(prevPt,bndPt,x);
		mt.mirror(x,n,x);
		mt.add(x,bndPt,x);
		return x;
	}
	public void calcStartIndex() {
		int pos = -1;
		for(int i = cells.size()-1;i>=0;i--) {
			short[] cell = cells.get(i);
			if(cell[0]>=0&&cell[1]>=0&&cell[0]<gridSize&&cell[1]<gridSize)
				pos = i;
			else
				break;
		}
		startIndex = pos;
	}
	@Override
	public int getPeriod() {
		if(isPropagating())
			return -1;
		int bndPts = startBndValType==POINT?1:0;
		bndPts += endBndValType==POINT?1:0;
		return getRotations()*(2*cells.size()-bndPts)/2;
	}
	/**
	 * Determines how many times the pattern sequences needs to be rotated
	 * for a complete period.
	 */
	private int getRotations() {
		if(isPropagating())
			return -1;
		byte rotations = 8;
		byte v1 = startBndVal.byteValue();
		byte v2 = endBndVal.byteValue();
		if(startBndValType!=POINT)
			v1 += 2;
		if(endBndValType!=POINT)
			v2 += 2;
		if(Math.abs(v1-v2)==4)
			rotations = 2;
		else if(v1%2==0==(v2%2==0))
			rotations = 4;
		return rotations;
	}
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("boundary_types = ");
		switch(getStartBndValType()) {
		case BaseLightPattern.POINT:
			sb.append("p");
			break;
		case BaseLightPattern.GRAD_LEFT:
			sb.append("-g");
			break;
		case BaseLightPattern.GRAD_RIGHT:
			sb.append("+g");
		}
		sb.append(",");
		switch(getEndBndValType()) {
		case BaseLightPattern.POINT:
			sb.append("p");
			break;
		case BaseLightPattern.GRAD_LEFT:
			sb.append("-g");
			break;
		case BaseLightPattern.GRAD_RIGHT:
			sb.append("+g");
		}
		sb.append("\n");
		sb.append("boundary_values = ");
		sb.append(getStartBndVal()).append(",");
		sb.append(getEndBndVal()).append("\n");
		sb.append("initial_gradients = ");
		for(float ig:getInitialGradients()) {
			String val = ig+"";
			int idx = val.indexOf(".");
			if(idx!=-1)
				val = val.substring(0,Math.min(idx+3,val.length()));
			sb.append(val).append(",");
		}
		sb.replace(sb.length()-1,sb.length(),"\n");
		sb.append("reflection_path = ");
		for(short[] cell:getCellSequence()) {
			sb.append("(").append(cell[0]).append(",");
			sb.append(cell[1]).append("),");
		}
		sb.replace(sb.length()-1,sb.length(),"\n");
		sb.append("grid_size = ");
		sb.append(getGridSize()).append("\n");
		sb.append("max_radius = ");
		T maxr = getMaxShapeSize();
		sb.append(mt.util.toPlainString(maxr)).append("%\n");
		sb.append("store_calc = ");
		sb.append(isStoreCalcEnabled());
		return sb.toString();
	}
	public void setLogging(boolean l) {
		this.log = l;
	}
}
